from npl.optimization.local_optimization.local_optimization import local_optimization

import copy


class LocalOptimizationOperator:
    def __init__(self, energy_calculator, environment_energies):
        self.energy_calculator = energy_calculator
        self.environment_energies = environment_energies

    def local_optimization(self, particle):
        new_particle = copy.deepcopy(particle)
        new_particle, accepted_energies = local_optimization(new_particle,
                                                             self.energy_calculator,
                                                             self.energy_calculator)

        return new_particle, accepted_energies



import numpy as np
from sklearn.cluster import KMeans
import copy


class LocalEnvironmentFeatureClassifier:
    """
    LocalEnvironmentFeatureClassifier is a class that classifies features of local environments around atoms in a particle.

    Attributes:
        local_environment_calculator: An instance responsible for computing local environments.
        feature_key: A key to identify the feature being computed.

    Methods:
        __init__(local_environment_calculator):
            Initializes the classifier with a local environment calculator.

        compute_atom_features(particle, recompute_local_environments=False):
            Computes features for each atom in the particle. Optionally recomputes local environments.

        compute_feature_vector(particle, recompute_atom_features=True, recompute_local_environments=False):
            Computes the feature vector for the particle. Optionally recomputes atom features and local environments.

        compute_atom_feature(particle, atom_index, recompute_local_environment=False):
            Computes the feature for a specific atom in the particle.

        get_feature_key():
            Returns the feature key.

        set_feature_key(feature_key):
            Sets the feature key.

        compute_n_features(particle):
            Abstract method to compute the number of features for the particle. Must be implemented by subclasses.

        predict_atom_feature(particle, lattice_index, recompute_local_environment=False):
            Abstract method to predict the feature of a specific atom. Must be implemented by subclasses.
    """
    def __init__(self, local_environment_calculator):
        self.local_environment_calculator = local_environment_calculator
        self.feature_key = None

    def compute_atom_features(self, particle, recompute_local_environments=False):
        if recompute_local_environments:
            self.local_environment_calculator.compute_local_environments(particle)

        for atom_index in particle.get_indices():
            self.compute_atom_feature(particle, atom_index, recompute_local_environments)

    def compute_feature_vector(self, particle, recompute_atom_features=True, recompute_local_environments=False):
        if recompute_atom_features:
            self.compute_atom_features(particle, recompute_local_environments)

        n_features = self.compute_n_features(particle)
        feature_vector = np.zeros(n_features)
        atom_features = particle.get_atom_features(self.feature_key)
        for index in particle.get_indices():
            feature_vector[atom_features[index]] += 1

        particle.set_feature_vector(self.feature_key, feature_vector)

    def compute_atom_feature(self, particle, atom_index, recompute_local_environment=False):
        feature = self.predict_atom_feature(particle, atom_index, recompute_local_environment)
        atom_features = particle.get_atom_features(self.feature_key)
        atom_features[atom_index] = feature

    def get_feature_key(self):
        return self.feature_key

    def set_feature_key(self, feature_key):
        self.feature_key = feature_key

    def compute_n_features(self, particle):
        raise NotImplementedError

    def predict_atom_feature(self, particle, lattice_index, recompute_local_environment=False):
        raise NotImplementedError


class KMeansClassifier(LocalEnvironmentFeatureClassifier):
    def __init__(self, n_cluster, local_environment_calculator, feature_key):
        LocalEnvironmentFeatureClassifier.__init__(self, local_environment_calculator)
        self.kMeans = None
        self.n_cluster = n_cluster
        self.feature_key = feature_key

    def compute_n_features(self, particle):
        n_elements = len(particle.get_all_symbols())
        n_features = self.n_cluster * n_elements
        return n_features

    # TODO problematic: length of feature vector different for pure particles and bimetallic ones
    def predict_atom_feature(self, particle, atom_index, recompute_local_environment=False):
        symbol = particle.get_symbol(atom_index)
        symbols = sorted(particle.get_all_symbols())
        symbol_index = symbols.index(symbol)

        offset = symbol_index*self.n_cluster
        if recompute_local_environment:
            environment = self.kMeans.predict([self.local_environment_calculator.predict_local_environment(particle, atom_index)])[0]
        else:
            environment = self.kMeans.predict([particle.get_local_environment(atom_index)])[0]
        return offset + environment

    def kmeans_clustering(self, training_set):
        local_environments = list()
        for particle in training_set:
            local_environments = local_environments + list(particle.get_local_environments().values())

        print("Starting kMeans")
        self.kMeans = KMeans(n_clusters=self.n_cluster, random_state=0).fit(local_environments)


# TODO rename, rework with offsets etc.
class TopologicalEnvironmentClassifier(LocalEnvironmentFeatureClassifier):
    def __init__(self, local_environment_calculator, symbols):
        LocalEnvironmentFeatureClassifier.__init__(self, local_environment_calculator)
        symbols_copy = copy.deepcopy(symbols)
        symbols_copy.sort()
        self.symbols = symbols_copy

        self.coordination_number_offsets = [int(cn*(cn + 1)/2) for cn in range(13)]

        self.feature_key = 'TEC'

    def compute_n_features(self, particle):
        return 182

    def predict_atom_feature(self, particle, atom_index, recompute_local_environment=False):
        symbol = particle.get_symbol(atom_index)
        symbol_index = self.symbols.index(symbol)

        element_offset = symbol_index*91

        if recompute_local_environment:
            self.local_environment_calculator.compute_local_environment(particle, atom_index)

        environment = particle.get_local_environment(atom_index)
        coordination_number = len(particle.neighbor_list[atom_index])

        atom_feature = int(element_offset + self.coordination_number_offsets[coordination_number] + environment[0])

        return atom_feature


# TODO rename
class CoordinationNumberClassifier(LocalEnvironmentFeatureClassifier):
    def __init__(self, local_environment_calculator):
        LocalEnvironmentFeatureClassifier.__init__(self, local_environment_calculator)
        self.coordination_number_offsets = [int(cn * (cn + 1) / 2) for cn in range(13)]

        self.feature_key = 'TEC'

    def compute_n_features(self, particle):
        return 182

    def predict_atom_feature(self, particle, atom_index, recompute_local_environment=False):
        symbol = particle.get_symbol(atom_index)
        if symbol == 'X':
            symbol_index = 1
        else:
            symbol_index = 0

        element_offset = symbol_index*91

        if recompute_local_environment:
            self.local_environment_calculator.compute_local_environment(particle, atom_index)

        environment = particle.get_local_environment(atom_index)
        coordination_number = environment[0]  # TODO not robust, only works if 'X' as 'empty site' is second entry
        # TODO should specify index of non-vacancy element

        atom_feature = element_offset + self.coordination_number_offsets[coordination_number] + environment[0]

        return atom_feature




from typing import Union

from ase import Atoms
from ase.optimize import BFGS
from ase.calculators.calculator import Calculator
import pickle

class TOPCalculator(Calculator):
    """
    A class representing a calculator for performing relaxation calculations using the ASE library.
    
    Parameters:
        calculator (Calculator): The calculator object used for performing the calculations.
        fmax (float): The maximum force tolerance for the relaxation.
    """
    
    def __init__(self,
                 model_paths : Union[list, str],
                 feature_key : str,
                 **kwargs
                 ):
        Calculator.__init__(self, **kwargs)

        self.feature_key = feature_key
        self.model = self.load_model(model_paths)
    
    def load_model(self, model_path):
        with open(model_path, 'rb') as calc:
            return pickle.load(calc)
    
    def calculate(self, atoms):
        self.results = {}
        feature_vector = atoms.info[self.feature_key]
        self.results['energy'] = self.model.predict(feature_vector)



import numpy as np

def get_atom_position(atom):
    return atom.position

def get_dot_product(v1, v2):
    return np.dot(v1,v2)

def get_unit_vector(position):
    length = np.linalg.norm(position)
    unit_vector = position / length

    return unit_vector, length

def get_perpendicular_vector(v, tolerance = 1e-04):
    if abs(v[1]) < tolerance and abs(v[2]) < tolerance:
        if abs(v[0]) < tolerance:
            raise ValueError('zero vector')
        else:
            return np.cross(v, [0, 1, 0])
    return np.cross(v, [1, 0 , 0])

def find_middle_point(positions):
    return sum(positions)/len(positions)    

def get_perpendicular_edge_vector(position, tolerance=1e-4):
    final_vector = np.array([0,0,0])

    for i, coordinate in enumerate(position):
        trial_vector = np.array([0,0,0])
        if abs(coordinate) > tolerance:
            trial_vector[i] = coordinate/abs(coordinate)
            final_vector += trial_vector
        else:
            final_vector += trial_vector
    final_vector, _ = get_unit_vector(final_vector)
    return final_vector

def get_normal_vector(positions):
    v1 = positions[1] - positions[0]
    v2 = positions[2] - positions[0]
    normal = np.cross(v1, v2)
    unit, _ = get_unit_vector(normal)
    return unit

def get_plane(positions):
    normal_vector = get_normal_vector(positions)
    d = np.dot(normal_vector, positions[2])
    return normal_vector, d

def get_bridge_perpendicular_line(positions, center_of_mass):
    a, a1 = positions
    n,_ = get_unit_vector(a - a1)
    p = center_of_mass
    direction = (p - a) - (np.dot((p-a), n)) * n
    return direction / np.linalg.norm(direction)

    



import unittest
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

class TestNotebooks(unittest.TestCase):
    def test_adsorbate_distribution(self):
        with open('Examples/adsorbate_distribution.ipynb') as f:
            nb = nbformat.read(f, as_version=4)
            ep = ExecutePreprocessor(timeout=600, kernel_name='python3')
            ep.preprocess(nb, {'metadata': {'path': 'Examples/'}})

    def test_layer_descriptor(self):
        with open('Examples/layer_descriptor.ipynb') as f:
            nb = nbformat.read(f, as_version=4)
            ep = ExecutePreprocessor(timeout=600, kernel_name='python3')
            ep.preprocess(nb, {'metadata': {'path': 'Examples/'}})

    # Add similar methods for other notebooks

if __name__ == '__main__':
    unittest.main()



import numpy as np
import copy
from itertools import chain

from npl.descriptors.local_environment_calculator import NeighborCountingEnvironmentCalculator
from npl.monte_carlo.random_exchange_operator import RandomExchangeOperator

import time

def setup_monte_carlo(start_particle, energy_calculator, local_feature_classifier):
    symbols = start_particle.get_all_symbols()
    energy_key = energy_calculator.get_energy_key()

    local_env_calculator = NeighborCountingEnvironmentCalculator(symbols)
    local_env_calculator.compute_local_environments(start_particle)

    local_feature_classifier.compute_feature_vector(start_particle)
    energy_calculator.compute_energy(start_particle)

    exchange_operator = RandomExchangeOperator(0.5)
    exchange_operator.bind_particle(start_particle)

    return energy_key, local_env_calculator, exchange_operator


def update_atomic_features(exchanges, local_env_calculator, local_feature_classifier, particle):
    neighborhood = set()
    for exchange in exchanges:
        index1, index2 = exchange
        neighborhood.add(index1)
        neighborhood.add(index2)

        neighborhood = neighborhood.union(particle.neighbor_list[index1])
        neighborhood = neighborhood.union(particle.neighbor_list[index2])

    for index in neighborhood:
        local_env_calculator.compute_local_environment(particle, index)
        local_feature_classifier.compute_atom_feature(particle, index)

    local_feature_classifier.compute_feature_vector(particle, recompute_atom_features=False)
    return particle, neighborhood

def run_monte_carlo(beta, max_steps, start_particle, energy_calculator, local_feature_classifier):
    energy_key, local_env_calculator, exchange_operator = setup_monte_carlo(start_particle, energy_calculator,
                                                                            local_feature_classifier)

    start_energy = start_particle.get_energy(energy_key)
    lowest_energy = start_energy
    accepted_energies = [(lowest_energy, 0)]

    found_new_solution = False
    fields = ['energies', 'symbols']
    best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))

    total_steps = 0
    no_improvement = 0
    while no_improvement < max_steps:
        total_steps += 1
        if total_steps % 2000 == 0:
            print("Step: {}".format(total_steps))
            print("Lowest energy: {}".format(lowest_energy))


        exchanges = exchange_operator.random_exchange(start_particle)

        start_particle, neighborhood = update_atomic_features(exchanges, local_env_calculator, local_feature_classifier,
                                                              start_particle)

        
        energy_calculator.compute_energy(start_particle)
        new_energy = start_particle.get_energy(energy_key)

        delta_e = new_energy - start_energy

        acceptance_rate = min(1, np.exp(-beta * delta_e))
        if np.random.random() < acceptance_rate:
            if found_new_solution:
                if new_energy > start_energy:
                    start_particle.swap_symbols(exchanges)
                    best_particle = copy.deepcopy(start_particle)
                    start_particle.swap_symbols(exchanges)

            start_energy = new_energy
            accepted_energies.append((new_energy, total_steps))

            if new_energy < lowest_energy:
                no_improvement = 0
                lowest_energy = new_energy
                found_new_solution = True
            else:
                no_improvement += 1
                found_new_solution = False

        else:
            no_improvement += 1

            # roll back exchanges and make sure features and environments are up-to-date
            start_particle.swap_symbols(exchanges)
            start_particle.set_energy(energy_key, start_energy)
            for index in neighborhood:
                local_env_calculator.compute_local_environment(start_particle, index)
                local_feature_classifier.compute_atom_feature(start_particle, index)

            if found_new_solution:
                best_particle = copy.deepcopy(start_particle)
                found_new_solution = False

    accepted_energies.append((accepted_energies[-1][0], total_steps))

    return [best_particle, accepted_energies]

def run_monte_carlo_for_adsorbates(beta, max_steps, start_particle, adsorbates_energy, n_adsorbates):
    #from Core.GlobalFeatureClassifier import AdsorptionFeatureVector
    #from Core.EnergyCalculator import LateralInteractionCalculator

    exchange_operator = RandomExchangeOperator(0.5)
    exchange_operator.bind_adsorbates(start_particle, n_adsorbates)

    symbols = start_particle.get_all_symbols()
    adsorbates_energy_key = adsorbates_energy.get_energy_key()

    global_feature_classifier = AdsorptionFeatureVector(symbols)
    global_feature_classifier.compute_feature_vector(start_particle)
    adsorbates_energy.compute_energy(start_particle)

    lateral_interaction_calculator = LateralInteractionCalculator()
    lateral_interaction_energy_key = lateral_interaction_calculator.energy_key
    lateral_interaction_calculator.bind_grid(start_particle)
    lateral_interaction_calculator.compute_energy(start_particle)

    def get_ordering_and_adsorbates_energy(particle):
        adsorbates_energy = start_particle.get_energy(adsorbates_energy_key)
        lateral_interaction = start_particle.get_energy(lateral_interaction_energy_key)
        particle.set_energy('TOT', adsorbates_energy+lateral_interaction)
        return particle.get_energy('TOT')
    #energy_key, local_env_calculator, exchange_operator = setup_monte_carlo(start_particle, energy_calculator, local_feature_classifier)


    #initial_adsorbed_sites = start_particle.get_indices_of_adsorbates()
    

    start_energy = get_ordering_and_adsorbates_energy(start_particle)
    lowest_energy = start_energy
    sites_occupied = start_particle.get_indices_of_adsorbates()
    accepted_energies = [(lowest_energy, 0, sites_occupied)]

    found_new_solution = False
    fields = ['energies', 'symbols', 'positions']
    best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))

    total_steps = 0
    no_improvement = 0
    while no_improvement < max_steps:
        total_steps += 1
        if total_steps % 2000 == 0:
            print("Step: {}".format(total_steps))
            print("Lowest energy: {}".format(lowest_energy))

        exchanges = exchange_operator.random_adsorbate_migration(start_particle)

        accepted_particle = copy.deepcopy(start_particle)

        global_feature_classifier.compute_feature_vector(start_particle)
        adsorbates_energy.compute_energy(start_particle)
        lateral_interaction_calculator.compute_energy(start_particle)
        new_energy = get_ordering_and_adsorbates_energy(start_particle)


        delta_e = new_energy - start_energy

        acceptance_rate = min(1, np.exp(-beta * delta_e))
        if np.random.random() < acceptance_rate:
            if found_new_solution:
                if new_energy > start_energy:
                    start_particle.swap_status(exchanges)
                    best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))
                    best_particle['energies'][adsorbates_energy_key] = copy.deepcopy(start_energy)
                    best_particle['ads'] = copy.deepcopy(start_particle.get_indices_of_adsorbates())
                    best_particle['lateral_interaction'] = start_particle.get_energy(lateral_interaction_energy_key)
                    best_particle['adsorbates_energy'] = start_particle.get_energy(adsorbates_energy_key)
                    start_particle.swap_status(exchanges)

            sites_occupied = start_particle.get_indices_of_adsorbates()
            start_energy = new_energy
            accepted_energies.append((new_energy, total_steps, sites_occupied))

            if new_energy < lowest_energy:
                no_improvement = 0
                lowest_energy = new_energy
                found_new_solution = True
            else:
                no_improvement += 1
                found_new_solution = False

        else:
            no_improvement += 1

            # roll back exchanges and make sure features and environments are up-to-date
            start_particle.swap_status(exchanges)
            start_particle.set_energy('TOT', start_energy)

            if found_new_solution:
                best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))
                best_particle['energies'][adsorbates_energy_key] = copy.deepcopy(start_energy)
                best_particle['ads'] = copy.deepcopy(start_particle.get_indices_of_adsorbates())
                best_particle['lateral_interaction'] = start_particle.get_energy(lateral_interaction_energy_key)
                best_particle['adsorbates_energy'] = start_particle.get_energy(adsorbates_energy_key)
                found_new_solution = False

    accepted_energies.append((accepted_energies[-1][0], total_steps))
    return [best_particle, accepted_energies]


def run_monte_carlo_ordering_adsorbates(beta, max_steps, start_particle, ordering_energy_calculator, adsorbates_energy_calculator, n_adsorbates, local_feature_classifier):
    from npl.descriptors.global_feature_classifier import AdsorptionFeatureVector
    from npl.calculators.energy_calculator import LateralInteractionCalculator

    exchange_operator = RandomExchangeOperator(0.5)
    exchange_operator.bind_adsorbates(start_particle, n_adsorbates)

    symbols = start_particle.get_all_symbols()
    adsorbates_energy_key = adsorbates_energy_calculator.get_energy_key()

    adsorbate_feature_classifier = AdsorptionFeatureVector(symbols)
    adsorbate_feature_classifier.compute_feature_vector(start_particle)
    adsorbates_energy_calculator.compute_energy(start_particle)

    ordering_energy_key, local_env_calculator, ordering_exchange_operator = setup_monte_carlo(start_particle, ordering_energy_calculator, local_feature_classifier)

    initial_adsorbed_sites = start_particle.get_indices_of_adsorbates()

    lateral_interaction_calculator = LateralInteractionCalculator()
    lateral_interaction_energy_key = lateral_interaction_calculator.energy_key
    lateral_interaction_calculator.bind_grid(start_particle)
    lateral_interaction_calculator.compute_energy(start_particle)

    def get_ordering_and_adsorbates_energy(particle):
        ordering_energy = start_particle.get_energy(ordering_energy_key)
        adsorbates_energy = start_particle.get_energy(adsorbates_energy_key)
        lateral_interaction = start_particle.get_energy(lateral_interaction_energy_key)
        particle.set_energy('TOT', ordering_energy + adsorbates_energy+lateral_interaction)
        return particle.get_energy('TOT')

    #start_energy = start_particle.get_energy(ordering_energy_key) + start_particle.get_energy(adsorbates_energy_key)
    start_energy = get_ordering_and_adsorbates_energy(start_particle)
    
    lowest_energy = start_energy
    sites_occupied = start_particle.get_indices_of_adsorbates()
    accepted_energies = [(lowest_energy, 0, sites_occupied)]

    found_new_solution = False
    fields = ['energies', 'symbols', 'positions']
    best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))
    best_particles = list()
    total_steps = 0
    no_improvement = 0
    while no_improvement < max_steps:
        total_steps += 1
        if total_steps % 2000 == 0:
            print("Step: {}".format(total_steps))
            print("Lowest energy: {}".format(lowest_energy))

        ordering_exchanges, adsorbates_exchanges = exchange_operator.coupled_random_exchange(start_particle)

        start_particle, neighborhood = update_atomic_features(ordering_exchanges, local_env_calculator, local_feature_classifier,
                                                              start_particle)

        accepted_particle = copy.deepcopy(start_particle)
        adsorbate_feature_classifier.compute_feature_vector(start_particle)

        ordering_energy_calculator.compute_energy(start_particle)
        adsorbates_energy_calculator.compute_energy(start_particle)
        lateral_interaction_calculator.compute_energy(start_particle)
        new_energy = get_ordering_and_adsorbates_energy(start_particle)

        delta_e = new_energy - start_energy

        acceptance_rate = min(1, np.exp(-beta * delta_e))
        if np.random.random() < acceptance_rate:
            if found_new_solution:
                if new_energy > start_energy:
                    start_particle.swap_symbols(ordering_exchanges)
                    start_particle.swap_status(adsorbates_exchanges)
                    best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))
                    best_particle['energies'][ordering_energy_key] = copy.deepcopy(start_energy)
                    best_particle['ads'] = copy.deepcopy(start_particle.get_indices_of_adsorbates())
                    best_particles.append(best_particle)
                    start_particle.swap_status(adsorbates_exchanges)
                    start_particle.swap_symbols(ordering_exchanges)

            sites_occupied = start_particle.get_indices_of_adsorbates()
            start_energy = new_energy
            accepted_energies.append((new_energy, total_steps, sites_occupied))

            if new_energy < lowest_energy:
                no_improvement = 0
                lowest_energy = new_energy
                found_new_solution = True
            else:
                no_improvement += 1
                found_new_solution = False

        else:
            no_improvement += 1

            # roll back exchanges and make sure features and environments are up-to-date
            start_particle.swap_status(adsorbates_exchanges)
            start_particle.swap_symbols(ordering_exchanges)
            start_particle.set_energy('TOT', start_energy)

            for index in neighborhood:
                local_env_calculator.compute_local_environment(start_particle, index)
                local_feature_classifier.compute_atom_feature(start_particle, index)

            if found_new_solution:
                best_particle = copy.deepcopy(start_particle.get_as_dictionary(fields))
                best_particle['energies'][ordering_energy_key] = copy.deepcopy(start_energy)
                best_particle['ads'] = copy.deepcopy(start_particle.get_indices_of_adsorbates())
                best_particles.append(best_particle)
                found_new_solution = False

    accepted_energies.append((accepted_energies[-1][0], total_steps))
    return [best_particle, accepted_energies, best_particles]



import numpy as np
from npl.optimization.genetic_algoritm.CutAndSpliceOperator import CutAndSpliceOperator
from npl.optimization.genetic_algoritm.ExchangeOperator import ExchangeOperator
from npl.optimization.local_optimization.local_optimization import local_optimization


def compute_fitness(particle, min_energy, max_energy, energy_key):
    if max_energy == min_energy:
        return 0
    normalized_energy = (particle.get_energy(energy_key) - min_energy) / (max_energy - min_energy)
    return np.exp(-3 * normalized_energy)


def run_single_particle_ga(start_population, unsuccessful_gens_for_convergence, energy_calculator, local_env_calculator,
                           local_feature_classifier, environment_energies):
    unsuccessful_gens = 0
    energy_key = energy_calculator.get_energy_key()

    for p in start_population:
        local_env_calculator.compute_local_environments(p)
        local_feature_classifier.compute_feature_vector(p)
        energy_calculator.compute_energy(p)

    exchange_operator = ExchangeOperator(0.5)
    cut_and_splice_operator = CutAndSpliceOperator(10)

    cur_population = start_population
    generation = 0
    best_energies = []
    energy_evaluations = len(start_population)

    cur_population.sort(key=lambda x: x.get_energy(energy_key))
    best_energies.append((cur_population[0].get_energy(energy_key), energy_evaluations))
    while unsuccessful_gens < unsuccessful_gens_for_convergence:
        if generation % 1 == 0:
            print("Generation: {}".format(generation))
            print(energy_evaluations)

        min_energy = cur_population[0].get_energy(energy_key)
        max_energy = cur_population[-1].get_energy(energy_key)
        generation += 1

        fitness_values = np.array([compute_fitness(p, min_energy, max_energy, energy_key) for p in cur_population])
        if np.sum(fitness_values) == 0:
            break
        fitness_values /= np.sum(fitness_values)

        while True:
            p = np.random.rand()
            if p < 0.4:
                parent1, parent2 = np.random.choice(cur_population, 2, replace=False, p=fitness_values)
                new_offspring = cut_and_splice_operator.cut_and_splice(parent1, parent2)
            else:
                parent = np.random.choice(cur_population, 1, p=fitness_values)[0]
                new_offspring = exchange_operator.random_exchange(parent)

            new_offspring, energies = local_optimization(new_offspring,
                                                         energy_calculator,
                                                         environment_energies)

            energy_evaluations += energies[-1][1]

            new_energy = new_offspring.get_energy(energy_key)
            unique = True
            for particle in cur_population:
                if np.abs(new_energy - particle.get_energy(energy_key)) < 1e-6:
                    unique = False
                    break
            if unique:
                break

        cur_population.append(new_offspring)

        cur_population.sort(key=lambda x: x.get_energy(energy_key))
        cur_population = cur_population[:-1]

        if cur_population[0].get_energy(energy_key) == best_energies[-1][0]:
            unsuccessful_gens += 1
        else:
            unsuccessful_gens = 0
            best_energies.append((cur_population[0].get_energy(energy_key), energy_evaluations))
            print("New best energy: {}".format(cur_population[0].get_energy(energy_key)))

    return [best_energies, cur_population[0], energy_evaluations]



import numpy as np
import copy


class MutationOperator:
    def __init__(self, p, symbols):
        self.p = p
        self.symbols = set(symbols)

    def random_mutation(self, particle, n_mutations=None, symbols=None):
        new_particle = copy.deepcopy(particle)
        if new_particle.is_pure():
            symbol_from = new_particle.atoms.get_all_symbols()[0]
            print(symbol_from)
            other_symbols = list(self.symbols.difference({symbol_from}))
            print(other_symbols)
            symbol_to = np.random.choice(other_symbols, 1)[0]
            print(symbol_to)
        else:
            if symbols is None:
                symbols = np.random.choice(new_particle.atoms.get_all_symbols(), 2, replace=False)
                symbol_from = symbols[0]
                symbol_to = symbols[1]
            else:
                symbol_from = symbols[0]
                symbol_to = symbols[1]

        max_mutations = min(new_particle.get_stoichiometry().values())
        if n_mutations is None:
            n_mutations = min(self.draw_from_geometric_distribution(), max_mutations)

        atom_indices_to_be_transformed = np.random.choice(new_particle.atoms.get_indices_by_symbol(symbol_from), n_mutations, replace=False)

        new_particle.atoms.transform_atoms(zip(atom_indices_to_be_transformed, [symbol_to] * n_mutations), a)

        return new_particle

    def draw_from_geometric_distribution(self):
        return np.random.geometric(p=self.p, size=1)[0]



import os
import random
import warnings
import logging

from abc import ABC, abstractmethod
from collections import OrderedDict
from math import gcd
from time import time
from typing import Any, Dict, List, Optional, Type, Union
from ase.io.trajectory import Trajectory

import numpy as np

from ase import Atoms
from ase.calculators.calculator import Calculator

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class BaseEnsemble(ABC):
    def __init__(self,
                 structure: Atoms,
                 calculator: Calculator,
                 user_tag: Optional[str] = None,
                 random_seed: Optional[int] = None,
                 trajectory_write_interval: Optional[int] = None,
                 traj_file: str = 'traj_test.traj', 
                 outfile: str = 'outfile.out',
                 outfile_write_interval: int = 10) -> None:
        """
        Base class for ensembles in Monte Carlo simulations.

        Args:
            structure (Atoms): The initial configuration of the system.
            calculator (Calculator): The calculator object used for energy calculations.
            user_tag (str, optional): A user-defined tag for the ensemble. Defaults to None.
            random_seed (int, optional): The random seed for the random number generator. Defaults to None.
            trajectory_write_interval (int, optional): The interval at which to write trajectory files. Defaults to None.
            traj_file (str, optional): The file to write trajectory data. Defaults to 'traj_test.traj'.
            outfile (str, optional): The file to write output data. Defaults to 'outfile.out'.
            outfile_write_interval (int, optional): The interval at which to write output files. Defaults to 10.
        """
        self._accepted_trials = 0
        self._step = 0

        # calculator and configuration
        self._structure = structure
        self._calculator = calculator
        self._user_tag = user_tag

        self._traj_file = traj_file
        self._outfile = outfile
        self._outfile_write_interval = outfile_write_interval
        
        # random number generator
        if random_seed is None:
            self._random_seed = random.randint(0, int(1e16))
        else:
            self._random_seed = random_seed
        random.seed(a=self._random_seed)

    @property
    def structure(self) -> Atoms:
        """ Current configuration (copy). """
        return self._structure
    
    @property
    def step(self) -> int:
        """ Current trial step counter. """
        return self._step
    
    def write_outfile(self, step: int, energy: float) -> None:
        """
        Write the step and energy to the output file.

        Args:
            step (int): The current step.
            energy (float): The energy value.
        """
        try:
            with open(self._outfile, 'a') as outfile:
                outfile.write(f'STEP: {step} ENERGY: {energy}\n')
        except IOError as e:
            logger.error(f"Error writing to file {self._outfile}: {e}")
    
    def write_traj_file(self, atoms: Atoms) -> None:
        """
        Write the trajectory file.

        Args:
            atoms (Atoms): The atomic configuration.
        """
        try:
            with Trajectory(self._traj_file, 'a') as traj:
                traj.write(atoms)
        except IOError as e:
            logger.error(f"Error writing to trajectory file {self._traj_file}: {e}")


# npl/descriptors/__init__.py

from .global_feature_classifier import (GlobalFeatureClassifier,
                                        testTopologicalFeatureClassifier,
                                        TopologicalFeatureClassifier,
                                        ExtendedTopologicalFeaturesClassifier,
                                        AtomicCoordinationTypes,
                                        CoordinationFeatureClassifier)

from .local_environment_feature_classifier import (LocalEnvironmentFeatureClassifier,
                                                   CoordinationNumberClassifier,
                                                   TopologicalEnvironmentClassifier)

from .local_environment_calculator import (LocalEnvironmentCalculator,
                                           NeighborCountingEnvironmentCalculator)

__all__ = [
    "LayererTopologicalDescriptors",
    "LocalEnvironmentCalculator"
]


import cProfile, pstats, io


def profile(fnc):
    def inner(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()
        retval = fnc(*args, **kwargs)
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
        return retval

    return inner



from npl.core import Nanoparticle
from npl.utils.cutting_plane_utilities import SphericalCuttingPlaneGenerator
import numpy as np


class CutAndSpliceOperator:
    def __init__(self, max_radius, recompute_neighbor_list=True, normal_dir=None):
        self.cutting_plane_generator = SphericalCuttingPlaneGenerator(max_radius, 0.0, 0.0, normal_dir)
        self.recompute_neighbor_list = recompute_neighbor_list

    def cut_and_splice(self, p1, p2, fixed_stoichiometry=True):
        self.cutting_plane_generator.set_center(p1.get_ase_atoms().get_center_of_mass())
        self.cutting_plane_generator.set_max_radius(np.max(p1.get_ase_atoms().get_positions() -
                                                           p1.get_ase_atoms().get_center_of_mass()))
        # Ensure that we indeed cut, i.e. take atoms from both particles
        while True:
            cutting_plane = self.cutting_plane_generator.generate_new_cutting_plane()
            atom_indices_in_positive_subspace, _ = cutting_plane.split_atom_indices(p1.get_ase_atoms())
            _, atom_indices_in_negative_subspace = cutting_plane.split_atom_indices(p2.get_ase_atoms())

            if np.sum(atom_indices_in_negative_subspace) > 0 and np.sum(atom_indices_in_positive_subspace) > 0:
                break

        new_particle = Nanoparticle()
        new_particle.add_atoms(p1.get_ase_atoms().copy()[atom_indices_in_positive_subspace], False)
        new_particle.add_atoms(p2.get_ase_atoms().copy()[atom_indices_in_negative_subspace], False)

        if fixed_stoichiometry is True:
            target_stoichiometry = p1.get_stoichiometry()
            new_particle.adjust_stoichiometry(target_stoichiometry)

        if self.recompute_neighbor_list:
            new_particle.construct_neighbor_list()

        return new_particle



import npl.utils.math_modules as math

from ase import Atoms

import copy
import itertools
import numpy as np

import itertools
from collections import defaultdict

class AdsorptionSiteList():

    def __init__(self):
        self.list = defaultdict(lambda : set())
        self.total_n_sites = 0
        self.occupation_vector = []

    def __getitem__(self, item):
        return self.list[item]

    def __setitem__(self, key, value):
        self.list[key] = value

    def construct(self, particle):
        adsorption_site_list = self.build_site_list(particle)
        for site_index, atom_indices in enumerate(adsorption_site_list):
            self.list[site_index] = set(atom_indices)

        self.total_n_sites = len(self.list)
        self.occupation_vector = np.array([0 for _ in range(self.total_n_sites)])

    def random_occupation(self, number_of_adsorbates):
        # Reset the occupation vector
        self.occupation_vector = np.array([0 for _ in range(self.total_n_sites)])

        occupied_sites_indices = np.random.choice(np.arange(self.total_n_sites), number_of_adsorbates, replace=False)
        for site_index in occupied_sites_indices:
                self.occupation_vector[site_index] = 1 

    def get_occupation_vector(self):
        return self.occupation_vector

    def get_total_number_of_sites(self):
        return self.total_n_sites

    def get_number_of_adsorbates(self):
        return len(self.get_occupation_status_by_indices(1))

    def get_site_atom_indices(self, index):
        return self.list[index]

    def get_occupation_status_by_indices(self, status):
        # status : occupied (1) or unoccupied (0) adsorption site
        return np.where(self.occupation_vector == status)[0]

    def occupation_vector(self):
        return self.occupation_vector

    def get_site_size(self, index):
        return len(self.list[index])

    def swap_status(self, index_pairs):
        for idx1, idx2 in index_pairs:
            self.occupation_vector[idx1], self.occupation_vector[idx2] = self.occupation_vector[idx2], self.occupation_vector[idx1]

    def build_site_list(self, particle):
        def find_plane_for_bridge_atoms(particle, indices):
            uncoordinated_atoms = set(particle.get_atom_indices_from_coordination_number(range(12)))
            shell_1 = set(particle.get_coordination_atoms(indices[0]))
            shell_2 = set(particle.get_coordination_atoms(indices[1]))
            shared_atoms = uncoordinated_atoms.intersection(shell_1.intersection(shell_2))
            return list(shared_atoms)

        atoms_in_surface = set(particle.get_atom_indices_from_coordination_number(range(10)))

        ontop_sites = []
        for atom in atoms_in_surface:
            ontop_sites.append([atom])

        bridge_sites = []
        for central_atom_index in atoms_in_surface:
            central_atom_nearest_neighbors = set(particle.get_coordination_atoms(central_atom_index))
            for nearest_neighbor in atoms_in_surface.intersection(central_atom_nearest_neighbors):
                pair = sorted([central_atom_index, nearest_neighbor])
                if pair not in bridge_sites:
                    if len(pair) == 2:
                        bridge_sites.append(pair)

        hollow_sites = []
        for pair in bridge_sites:
            for third_atom in find_plane_for_bridge_atoms(particle, pair):
                triplet = copy.copy(pair)
                triplet.append(third_atom)
                triplet = sorted(triplet)
                if triplet not in hollow_sites:
                    if len(triplet) == 3:
                        hollow_sites.append(triplet)


        four_fold_hollow_sites = []
        atoms_in_100 = set(particle.get_atom_indices_from_coordination_number([6,7,8]))
        sub_surface = set(particle.get_atom_indices_from_coordination_number([12]))
        four_fold_hollow_sites = []

        for atom_idx in sub_surface:
            sub_neigh = set(particle.get_coordination_atoms(atom_idx))
            hollow_site = sub_neigh.intersection(atoms_in_100)
            if len(hollow_site) == 4 and hollow_site not in four_fold_hollow_sites:
                four_fold_hollow_sites.append(hollow_site)


        return ontop_sites + bridge_sites + hollow_sites + four_fold_hollow_sites


        
class FindAdsorptionSites():
    """ Class that identify and place add atoms based on the Generalized coordination Numbers of the nanoparticles"""
    def __init__(self):
        self.ontop = []
        self.bridge_positions = []
        self.hollow_positions = []
    
    def get_ontop_sites(self, particle):
        for atom in particle.get_atom_indices_from_coordination_number(range(10)):
            self.ontop.append([atom])
            
    def get_bridge_sites(self, particle):
        shell_atoms = set(particle.get_atom_indices_from_coordination_number(range(10)))
        for central_atom_index in shell_atoms:
            central_atom_nearest_neighbors = set(particle.get_coordination_atoms(central_atom_index))
            for nearest_neighbor in shell_atoms.intersection(central_atom_nearest_neighbors):
                pair = sorted([central_atom_index, nearest_neighbor])
                if pair not in self.bridge_positions:
                    self.bridge_positions.append(pair)
                    
    def get_hollow_sites(self, particle):      
        for pair in self.bridge_positions:
            for third_atom in self.find_plane_for_bridge_atoms(particle, pair):
                triplet = copy.copy(pair)
                triplet.append(third_atom)
                triplet = sorted(triplet)
                if triplet not in self.hollow_positions:
                    self.hollow_positions.append(triplet)
    
    def find_plane_for_bridge_atoms(self, particle, indices):
        uncoordinated_atoms = particle.get_atom_indices_from_coordination_number(range(12))
        uncoordinated_atoms = set(uncoordinated_atoms)
        shell_1 = set(particle.get_coordination_atoms(indices[0]))
        shell_2 = set(particle.get_coordination_atoms(indices[1]))
        shared_atoms = uncoordinated_atoms.intersection(shell_1.intersection(shell_2))
        return list(shared_atoms)


    def find_atom_plane_vec(self, particle, atom_idx):
        normal_vector = -1
        pos_vec = particle.get_position(atom_idx)/ np.linalg.norm(particle.get_position(atom_idx))
        planes = [[1,1,1], [-1,1,1], [1,-1,1], [1,1,-1], [-1,-1,1], [-1,1,-1], [1,-1,-1], [-1,-1,-1]]
        planes += [[1,0,0], [-1,0,0], [0,1,0], [0,-1,0], [0,0,1], [0,0,-1]]
        for plane in planes:
            mi_vec = plane / np.linalg.norm(plane)
            dot_prod = abs(np.dot(mi_vec, pos_vec))
            if dot_prod > normal_vector:
                normal_vector = dot_prod
                direction = copy.copy(mi_vec)
        return direction

    def find_direction_for_edges(self, particle, atom_idx, center_of_mass):
        positions = [particle.get_position(atom_idx)]
        for neighbor in particle.get_coordination_atoms(atom_idx):
            if particle.get_coordination_number(neighbor) < 7:
                positions.append(particle.get_position(neighbor))
                break
        if len(positions) == 1:
            for neighbor in particle.get_coordination_atoms(atom_idx):
                if particle.get_coordination_number(neighbor) == 7:
                    positions.append(particle.get_position(neighbor))
                    break
        direction = math.get_bridge_perpendicular_line(positions, center_of_mass)
        return direction

class PlaceAddAtoms():
    """Class that plance add atoms on positions identified by FindAdsorptionSites"""
    def __init__(self, symbols):
        self.adsorption_sites = FindAdsorptionSites()
        self.symbols = sorted(symbols)
        self.sites_list = []
        self.ontop_positions = {site[0] : [] for site in itertools.combinations_with_replacement(self.symbols,1)}
        self.bridge_positions = {''.join(list(site)) : [] for site in itertools.combinations_with_replacement(self.symbols,2)}
        self.hollow_positions = {''.join(list(site)) : [] for site in itertools.combinations_with_replacement(self.symbols,3)}

    def bind_particle(self, particle):
        self.com = particle.atoms.atoms.get_center_of_mass()
        particle.atoms.atoms.translate(-self.com)
        self.adsorption_sites.get_ontop_sites(particle)
        self.adsorption_sites.get_bridge_sites(particle)
        self.adsorption_sites.get_hollow_sites(particle)  
        self.get_total_adsorption_sites(particle)
        self.get_ontop_sites(particle)
        self.get_bridge_sties(particle)
        self.get_hollow_sties(particle)
    
    def get_ontop_sites(self, particle):
        for atom in self.adsorption_sites.ontop:
            self.ontop_positions[particle.get_symbol(atom[0])].append(atom[0])

    def get_bridge_sties(self, particle):
        for pairs in self.adsorption_sites.bridge_positions:
            self.bridge_positions[''.join(sorted(particle.get_symbols(pairs)))].append(pairs)
 
    def get_hollow_sties(self, particle):
        for triplet in self.adsorption_sites.hollow_positions:
            self.hollow_positions[''.join(sorted(particle.get_symbols(triplet)))].append(triplet)
            
    def get_total_adsorption_sites(self, particle):
        #self.sites_list += self.adsorption_sites.ontop
        #self.sites_list += self.adsorption_sites.bridge_positions
        self.sites_list += self.adsorption_sites.hollow_positions

    def get_xyz_site_from_atom_indices(self, particle, site):
        #if isinstance(site, (np.ndarray, np.generic)) or isinstance(site, int):
        site = list(site)
        if len(site) == 1:
            pos_vec = particle.get_position(site[0])
            cn = particle.get_coordination_number(site[0])
            if cn > 7:
                plane_direction = self.adsorption_sites.find_atom_plane_vec(particle, site[0])
                dot_prod = np.dot(pos_vec, plane_direction)
                direction = dot_prod/abs(dot_prod)
                xyz_site = (particle.get_position(site[0]))+(direction*plane_direction*2)
                return xyz_site
            if cn == 7:
                plane_direction = self.adsorption_sites.find_direction_for_edges(particle, site[0], self.com)
                dot_prod = np.dot(pos_vec, plane_direction)
                direction = dot_prod/abs(dot_prod)
                xyz_site = (particle.get_position(site[0]))+(direction*plane_direction*2)
                return xyz_site
            if cn < 7:
                unit_vector1, length1 = math.get_unit_vector(particle.get_position(site[0]))
                xyz_site = (unit_vector1*(length1+2))
                return xyz_site
        else:
            xyz_atoms = [particle.get_position(atom_index) for atom_index in site]
            xyz_site_plane = math.find_middle_point(xyz_atoms)
            unit_vector1, length1 = math.get_unit_vector(xyz_site_plane)
            if len(site) >= 3:
                normal_vector = math.get_normal_vector(xyz_atoms) 
            if len(site) == 2:
                cn1 = particle.get_coordination_number(site[0])
                cn2 = particle.get_coordination_number(site[1])
                shared_atoms = set(particle.get_coordination_atoms(site[0])).intersection(set(particle.get_coordination_atoms(site[1])))
                corner = set(particle.get_atom_indices_from_coordination_number([4,6])).intersection(shared_atoms)
                if cn1 < 9 and  cn2 < 9 and len(corner) == 0:    
                    positions = [particle.get_position(x) for x in site]
                    normal_vector = math.get_bridge_perpendicular_line(positions, self.com)
                else:    
                    third_atom = self.adsorption_sites.find_plane_for_bridge_atoms(particle, site)
                    if isinstance(third_atom, int):
                        xyz_third_atom = particle.get_position(third_atom)
                    else:
                        xyz_third_atom = particle.get_position(third_atom[0])
                    xyz_atoms.append(xyz_third_atom)
                    normal_vector = math.get_normal_vector(xyz_atoms)
                
            unit_vector2, length2 = math.get_unit_vector(normal_vector)
            dot_prod = np.dot(unit_vector1, unit_vector2)
            direction = dot_prod/abs(dot_prod)
            xyz_site = (unit_vector1*(length1))+(direction*unit_vector2*(1.4))
   
            return xyz_site

    def place_add_atom(self, particle, adsorbates, sites):
        add_atom_list = []
        for site in sites:
            xyz_site = self.get_xyz_site_from_atom_indices(particle, site)

            for adsorbate in adsorbates:
                add_atom = Atoms(adsorbate)   
                add_atom.translate(xyz_site)
                add_atom_list.append(add_atom)

                unit_vector1, length1 = math.get_unit_vector(xyz_site)
                xyz_site = (unit_vector1*(length1+1.18))
            
        for add_atom in add_atom_list:
            particle.add_atoms(add_atom, recompute_neighbor_list=False)
            
        return particle    








import copy
import itertools
from collections import defaultdict
from itertools import combinations_with_replacement

import numpy as np
from ase.data import atomic_numbers


class GlobalFeatureClassifier:
    """Base class for classifiers that produce a feature vector directly from a particle without the need
    for precomputing local environments.

    Valid implementations have to implement the compute_feature_vector(particle) method.
    """
    def __init__(self):
        self.feature_key = None

    def get_feature_key(self):
        return self.feature_key

    def set_feature_key(self, feature_key):
        self.feature_key = feature_key

    def compute_feature_vector(self, particle):
        raise NotImplementedError


# TODO better name
# TODO refactor handling of symbols
class SimpleFeatureClassifier(GlobalFeatureClassifier):
    """Base class for classifiers which provides handling of two elements and calculation of bond counts.

    Entries for different elements in feature vectors have to be ordered consistently. This class uses an
    alphabetical ordering of TWO elements. The returned feature vector consists of [n_aa_bonds/n_atoms,
    n_ab_bonds/n_atoms, n_bb_bonds/n_atoms, n_a_atoms*0.1].

    Args:
        symbols (list): A list of symbols representing the elements.

    Attributes:
        symbol_a (str): The first symbol in the ordered list of symbols.
        symbol_b (str): The second symbol in the ordered list of symbols.
        feature_key (str): The key used to store the feature vector in the particle object.

    """
    def __init__(self, symbols):
        GlobalFeatureClassifier.__init__(self)
        symbols_copy = copy.deepcopy(symbols)
        symbols_copy.sort()
        self.symbol_a = symbols_copy[0]
        if len(symbols) == 2:
            self.symbol_b = symbols_copy[1]

        self.feature_key = 'SFC'
        return

    def compute_feature_vector(self, particle):
        """Compute the feature vector for the given particle.

        Args:
            particle (Particle): The particle object for which to compute the feature vector.

        """
        n_aa_bonds, n_bb_bonds, n_ab_bonds = self.compute_respective_bond_counts(particle)
        n_atoms = particle.atoms.get_n_atoms()

        M = particle.get_stoichiometry()[self.symbol_a] * 0.1
        particle.set_feature_vector(self.feature_key, np.array([n_aa_bonds / n_atoms, n_bb_bonds / n_atoms, n_ab_bonds / n_atoms, M]))

    def compute_respective_bond_counts(self, particle):
        """Compute the respective bond counts for the given particle.

        Args:
            particle (Particle): The particle object for which to compute the bond counts.

        Returns:
            tuple: A tuple containing the counts of AA bonds, BB bonds, and AB bonds.

        """
        n_aa_bonds = 0
        n_ab_bonds = 0
        n_bb_bonds = 0

        for lattice_index_with_symbol_a in particle.atoms.get_indices_by_symbol(self.symbol_a):
            neighbor_list = particle.neighbor_list[lattice_index_with_symbol_a]
            for neighbor in neighbor_list:
                symbol_neighbor = particle.atoms.get_symbol(neighbor)

                if self.symbol_a != symbol_neighbor:
                    n_ab_bonds += 0.5
                else:
                    n_aa_bonds += 0.5

        for lattice_index_with_symbol_b in particle.atoms.get_indices_by_symbol(self.symbol_b):
            neighbor_list = particle.neighbor_list[lattice_index_with_symbol_b]
            for neighbor in neighbor_list:
                symbol_neighbor = particle.atoms.get_symbol(neighbor)

                if self.symbol_b == symbol_neighbor:
                    n_bb_bonds += 0.5
                else:
                    n_ab_bonds += 0.5

        return n_aa_bonds, n_bb_bonds, n_ab_bonds

class testTopologicalFeatureClassifier(SimpleFeatureClassifier):
    """Classifier for a generalization of the topological descriptors by Kozlov et al. (2015).
    Implemented for TWO elements, which are sorted alphabetically.The returned feature vector will have the form
    [n_aa_bonds/n_atoms, n_ab_bonds/n_atoms, n_bb_bonds/n_atoms, n_a_atoms*0.1, n_a(cn=0), n_a(cn=1), ..., n_a(cn=12].
    """
    def __init__(self, symbols):
        SimpleFeatureClassifier.__init__(self, symbols)
        self.feature_key = 'TFC'
        
    def get_feature_labels(self):
        """
        Generate a list of feature labels for the topological feature vector.

        The feature labels include:
        - Bond counts for AA, BB, and AB bonds.
        - Number of atoms of type A.
        - Coordination numbers for atoms of type A from 0 to 12.

        Returns:
            list: A list of feature labels.
        """
        bonds = [f'{self.symbol_a}{self.symbol_a}', f'{self.symbol_b}{self.symbol_b}', f'{self.symbol_a}{self.symbol_b}']
        n_symbol_a_atoms = [f'n_{self.symbol_a}']
        coordination_a = [f'{self.symbol_a}(cn={i})' for i in range(13)]
        return bonds + n_symbol_a_atoms + coordination_a
    
    def compute_feature_vector(self, particle):
        n_atoms = particle.get_n_atoms()
        n_aa_bonds, n_bb_bonds, n_ab_bonds = self.compute_respective_bond_counts(particle)
        coordinated_atoms = [len(particle.get_atom_indices_from_coordination_number([cn], symbol=self.symbol_a)) for cn in range(13)]

        M = particle.get_stoichiometry()[self.symbol_a]*0.1

        feature_vector = np.array([n_aa_bonds/n_atoms, n_bb_bonds/n_atoms, n_ab_bonds/n_atoms, M] + coordinated_atoms)
        particle.set_feature_vector(self.feature_key, feature_vector)


class ExtendedTopologicalFeaturesClassifier(GlobalFeatureClassifier):
    """
    Extention of the Topological Feature Classifier Class.
    Computes the topological feature vector for more than 2 metals in the nanoparticle
    """
    def __init__(self, symbols):
        GlobalFeatureClassifier.__init__(self)
        symbols_copy = copy.deepcopy(symbols)
        self.symbols = sorted(symbols_copy)
        self.number_of_species = len(self.symbols)
        self.bond_types = dict()
        self.layer_types = dict()
        self.n_sublayers = len(self.symbols)
        self.n_bond_features = 0
        self.n_features = 0
        self.sublayer_indices = []
        self.feature_key = 'ETOP'

        self.get_bond_types()
        self.get_layer_types()
        self.get_number_of_feature()
        

    def get_bond_types(self):
        for i, bond_types in enumerate(combinations_with_replacement(self.symbols, 2)):
            self.bond_types[bond_types] = i
        self.n_bond_features = len(self.bond_types)

    def get_layer_types(self):
        for idx, symbol in enumerate(self.symbols):
            self.layer_types[symbol] = idx + len(self.bond_types)

    def get_number_of_feature(self):
        self.n_features = self.n_bond_features + self.n_sublayers + 13*self.number_of_species

    def get_coordination_index(self,particle,index, symbol):
        coordination = particle.get_coordination_number(index)
        cn_index = self.n_bond_features + self.n_sublayers + (coordination) + (13*self.symbols.index(symbol))
        return cn_index

    def get_sublayer_indices(self, particle):
        self.sublayer_indices = np.zeros(particle.get_n_atoms())
        for index in particle.get_indices():
            if len(particle.neighbor_list[index]) < 12:
                for sub_index in particle.neighbor_list[index]:
                    if len(particle.neighbor_list[sub_index]) == 12:
                        self.sublayer_indices[sub_index] = 1
        

    def compute_atom_feature(self, particle, index):

        atom_feature = np.zeros(self.n_features)
        element1 = particle.get_symbol(index)
        atom_feature[self.layer_types[element1]] = self.sublayer_indices[index]
        cn_index = self.get_coordination_index(particle, index, element1)

        for neigh_index in particle.neighbor_list[index]:
            element2 = particle.get_symbol(neigh_index)
            bond_type = sorted([element1,element2])
            index_feature = self.bond_types[tuple(bond_type)]
            atom_feature[index_feature] += 0.5

        atom_feature[cn_index] += 1
        
        return atom_feature



    def compute_atom_features(self, particle):
        self.get_sublayer_indices(particle)
        particle.set_atom_features(np.zeros((particle.get_n_atoms(),self.n_features)), self.feature_key) 
        for atom_idx in particle.get_indices():
            atom_feature = self.compute_atom_feature(particle, atom_idx)
            particle.set_atom_feature(self.feature_key, atom_idx, atom_feature)


    def compute_feature_vector(self, particle):

        self.compute_atom_features(particle)
        atom_features = particle.get_atom_features(self.feature_key)
        feature_vector = atom_features.sum(axis=0)
        particle.set_feature_vector(self.feature_key, feature_vector)
    
    def update_atom_feature(self, particle, index):
        
        atom_feature = self.compute_atom_feature(particle, index)
        particle.set_atom_feature(self.feature_key, index, atom_feature)

    def update_feature_vector(self, particle, neighborhood):

        old_atom_features = []
        feature_vector = particle.get_feature_vector(self.feature_key)
        
        change = 0
        for index in neighborhood:
            old_atom_feature = copy.deepcopy(particle.get_atom_feature(self.feature_key, index))
            old_atom_features.append(old_atom_feature)
            change -= old_atom_feature

            new_atom_feature = self.compute_atom_feature(particle, index)
            change += new_atom_feature
            particle.set_atom_feature(self.feature_key, index, new_atom_feature)

        feature_vector += change

        return old_atom_features, change

    def downgrade_feature_vector(self, particle, neighborhood, old_atom_features, change):
        feature_vector = particle.get_feature_vector(self.feature_key)
        
        for index, atom_feature in zip(neighborhood, old_atom_features):
            particle.set_atom_feature(self.feature_key, index, atom_feature)
        feature_vector -= change

        


class TopologicalFeatureClassifier(SimpleFeatureClassifier):
    """Classifier for a generalization of the topological descriptors by Kozlov et al. (2015).

    Implemented for TWO elements, which are sorted alphabetically.The returned feature vector will have the form
    [n_aa_bonds/n_atoms, n_ab_bonds/n_atoms, n_bb_bonds/n_atoms, n_a_atoms*0.1, n_a(cn=0), n_a(cn=1), ..., n_a(cn=12].
    """
    def __init__(self, symbols):
        SimpleFeatureClassifier.__init__(self, symbols)
        self.feature_key = 'TFC'
        self.bond_scaling_factor = 1
        
    def get_feature_labels(self):
        """
        Generate a list of feature labels for the topological feature vector.

        The feature labels include:
        - Bond counts for AA, BB, and AB bonds.
        - Number of atoms of type A.
        - Coordination numbers for atoms of type A and B from 0 to 12.

        Returns:
            list: A list of feature labels.
        """
        bonds = [f'{self.symbol_a}{self.symbol_a}', f'{self.symbol_b}{self.symbol_b}', f'{self.symbol_a}{self.symbol_b}']
        n_symbol_a_atoms = [f'n_{self.symbol_a}']
        coordination_a = [f'{self.symbol_a}(cn={i})' for i in range(13)]
        coordination_b = [f'{self.symbol_b}(cn={i})' for i in range(13)]
        return bonds + coordination_a + coordination_b

    def compute_feature_vector(self, particle):
        n_atoms = particle.get_n_atoms()
        n_aa_bonds, n_bb_bonds, n_ab_bonds = self.compute_respective_bond_counts(particle)
        
        coordinated_atoms_a = [len(particle.get_atom_indices_from_coordination_number([cn], symbol=self.symbol_a)) for cn in range(13)]
        coordinated_atoms_b = [len(particle.get_atom_indices_from_coordination_number([cn], symbol=self.symbol_b)) for cn in range(13)]

        feature_vector = np.array([n_aa_bonds*self.bond_scaling_factor, n_bb_bonds*self.bond_scaling_factor,
                                   n_ab_bonds*self.bond_scaling_factor] + coordinated_atoms_a + coordinated_atoms_b)
        
        particle.set_feature_vector(self.feature_key, feature_vector)


class AtomicCoordinationTypes(SimpleFeatureClassifier):
    def __init__(self, symbols):
        SimpleFeatureClassifier.__init__(self, symbols)
        symbols_copy = copy.deepcopy(symbols)
        self.symbols = sorted(symbols_copy)
        self.feature_key = 'ACT'
        
        self.coordination_number_offsets = [int(cn*(cn + 1)/2) for cn in range(13)]
    
    def compute_n_features(self, particle):
        return 182
    
    def compute_atom_feature(self, atom_index, particle):
        symbol = particle.get_symbol(atom_index)
        symbol_index = self.symbols.index(symbol)

        element_offset = symbol_index*91

        coordination_number = len(particle.neighbor_list[atom_index])
        symbols = [particle.get_symbol(neigh_index) for neigh_index in particle.neighbor_list[atom_index]] 
        if symbol == self.symbol_a:
            n_ab_bonds = symbols.count(self.symbol_a) # it was symbol_b
        else:
            n_ab_bonds = symbols.count(self.symbol_a) # it was symbol_a

        atoms_feature = int(self.coordination_number_offsets[coordination_number] + n_ab_bonds + element_offset)

        return atoms_feature

    def compute_feature_vector(self, particle):
        feature_vector = np.zeros(self.compute_n_features(particle))

        for atom_index in particle.get_indices():
            atom_feature = self.compute_atom_feature(atom_index, particle)
            feature_vector[atom_feature] += 1

        particle.set_feature_vector(self.feature_key, feature_vector)


# TODO move to separate modules & rename
class CoordinationFeatureClassifier(SimpleFeatureClassifier):
    """Feature classifier that counts how often each coordination number appears if vacancies (atoms of element 'X')
     are present in an otherwise PURE particle.

     Form of the feature vector:
     [n_aa_bonds, n_a(cn=0), n_a(cn=1), ... n_a(cn=12)]
     """
    def __init__(self, symbols):
        SimpleFeatureClassifier.__init__(self, symbols)
        self.feature_key = 'TFC'

    def compute_feature_vector(self, particle):
        n_aa_bonds, n_bb_bonds, n_ab_bonds = self.compute_respective_bond_counts(particle)
        symbol_a_features = [0]*13
        for index in particle.get_indices():
            if particle.get_symbol(index) != 'X':
                env = particle.get_local_environment(index)
                symbol_a_features[env[0]] += 1

        features = [n_aa_bonds] + symbol_a_features
        feature_vector = np.array(features)
        particle.set_feature_vector(self.feature_key, feature_vector)


class DipoleMomentCalculator(SimpleFeatureClassifier):
    
    def __init__(self, symbols):
        SimpleFeatureClassifier.__init__(self, symbols)
        self.feature_key = 'MU'

    def compute_feature_vector(self, particle, charges = [1, -1]):
    
        symbols = particle.get_all_symbols()
        fake_charges = {symbols[0] : charges[0], symbols[1] : charges[1]}
        partial_charges = [fake_charges[symbol] for symbol in particle.get_symbols()]

        dipole_moments = []
        environments = []
        for central_atom_idx in particle.get_atom_indices_from_coordination_number([12]):
            particle.translate_atoms_positions(particle.get_position(central_atom_idx))
            dipole_moment = 0
            for atom_idx in particle.get_coordination_atoms(central_atom_idx):
                dipole_moment += partial_charges[atom_idx] * particle.get_position(atom_idx)

            dipole_moments.append(np.linalg.norm(dipole_moment))
            environments.append(particle.get_coordination_atoms(central_atom_idx))

        feature_vector = np.average(dipole_moments)#/particle.get_n_atoms()

        particle.set_feature_vector(self.feature_key, feature_vector)

class AdsorptionFeatureVector(SimpleFeatureClassifier):

    def __init__(self, symbols):
        SimpleFeatureClassifier.__init__(self, symbols)
        self.feature_key = 'ADS'
        self.features_type = defaultdict(lambda : 0)
        self.n_features = 0

        self.get_features(symbols)


    def compute_feature_vector(self, particle):
        feature_vector = np.array([0 for _ in range(self.n_features)])
        for site_occupied in particle.get_occupation_status_by_indices(1):
            site_type = sorted(particle.get_symbols(list(particle.get_site_atom_indices(site_occupied))))
            site_type = tuple(site_type)
            index = self.features_type[site_type]
            feature_vector[index] += 1

        particle.set_feature_vector(self.feature_key, feature_vector)

    def get_features(self, symbols):
        symbols = sorted(symbols)
        index = 0
        for number_of_atom_in_site in range(1,5):
            for site_type in itertools.combinations_with_replacement(symbols,number_of_atom_in_site):
                self.features_type[site_type] = index
                index += 1
        self.n_features = len(self.features_type)

class GeneralizedCoordinationNumber(SimpleFeatureClassifier):

    """
    Class that only works for Monometallic Nanopaticles
    """

    def __init__(self, particle):

        particle.construct_adsorption_list()
        self.ads_list = particle.get_adsorption_as_list()
        self.all_gcn = list(particle.get_generalized_coordination_numbers(self.ads_list).keys())
        self.n_features = len(self.all_gcn)
        self.gcn_site_list = np.zeros(particle.get_total_number_of_sites())
        self.feature_key = 'GCN'
    
        for i, site_atom_indices in enumerate(self.ads_list):
            self.gcn_site_list[i] = particle.get_generalized_coordination_number(site_atom_indices)

    def compute_feature_vector(self, particle):
        feature_vector = np.zeros(self.n_features)
        occupied_site_indices = particle.get_indices_of_adsorbates()
        print(occupied_site_indices)
        for occupied_site_index in occupied_site_indices:
            index = self.all_gcn.index(self.gcn_site_list[occupied_site_index])
            feature_vector[index] += 1  

        particle.set_feature_vector(self.feature_key, feature_vector)
        

class LayererTopologicalDescriptors(SimpleFeatureClassifier):
    

    def __init__(self, symbols, particle):
        SimpleFeatureClassifier.__init__(self, symbols)
        self.feature_key = 'LTOP'       
        self.top =  TopologicalFeatureClassifier(symbols)
        self.layers = self.get_layer_indices(particle)
        self.n1, self.n2 = atomic_numbers[symbols[0]], atomic_numbers[symbols[1]]
    
    def compute_layers(self, particle):
        positions = particle.get_positions()
        x_layer, y_layer, z_layer = [np.unique(particle.positions[:,coord]) for coord in range(3)]
        return x_layer, y_layer, z_layer

    def get_layer_indices(self, particle):
        positions = particle.get_positions()
        x_layer, y_layer, z_layer = self.compute_layers(particle)

        x_layer_indices = {x : np.where(positions[:,0] == x)[0] for x in x_layer}
        y_layer_indices = {y : np.where(positions[:,1] == y)[0] for y in y_layer}
        z_layer_indices = {z : np.where(positions[:,2] == z)[0] for z in z_layer}

        return x_layer_indices, y_layer_indices, z_layer_indices

    def get_layers_occupacy(self, particle):
        x_layer_indices, y_layer_indices, z_layer_indices = self.layers

        layers_occupacy_x = 0
        layers_occupacy_y = 0
        layers_occupacy_z = 0

        for layer in x_layer_indices.values():
            n_a = np.where(particle.atoms.atoms.numbers[layer]==self.n1)[0]
            n_b = np.where(particle.atoms.atoms.numbers[layer]==self.n2)[0]
            layers_occupacy_x += abs(len(n_a) - len(n_b))

        for layer in y_layer_indices.values():
            n_a = np.where(particle.atoms.atoms.numbers[layer]==self.n1)[0]
            n_b = np.where(particle.atoms.atoms.numbers[layer]==self.n2)[0]
            layers_occupacy_y += abs(len(n_a) - len(n_b))

        for layer in z_layer_indices.values():
            n_a = np.where(particle.atoms.atoms.numbers[layer]==self.n1)[0]
            n_b = np.where(particle.atoms.atoms.numbers[layer]==self.n2)[0]
            layers_occupacy_z += abs(len(n_a) - len(n_b))

        return np.array([layers_occupacy_x, layers_occupacy_y, layers_occupacy_z])
        #return np.array([layers_occupacy_x, layers_occupacy_x, layers_occupacy_x])
        

    def compute_feature_vector(self, particle):
        self.top.compute_feature_vector(particle)
        layericity = np.array(self.get_layers_occupacy(particle))
        
        feature_vector = np.empty(32) # it should be 32
        feature_vector[:29] = particle.get_feature_vector(self.top.get_feature_key())
        feature_vector[29:] = layericity

        particle.set_feature_vector(self.feature_key, feature_vector)


import numpy as np


class CuttingPlane:
    """Class that represents a cutting plane via an anchor point and a normal."""

    def __init__(self, anchor, normal):
        self.anchor = anchor
        self.normal = normal

    def split_atom_indices(self, atoms):
        """Split the given atoms into the positive and negative subspace of the plane.

        Return numpy arrays of type boolean that are True if the atom with the corresponding
        index lays within the respective subspace.

        Parameters:
            atoms : Atoms
        """
        dot_product = np.dot((atoms.positions - self.anchor), self.normal)
        indices_in_positive_subspace = dot_product > 0
        indices_in_negative_subspace = dot_product < 0
        return indices_in_positive_subspace, indices_in_negative_subspace


class SphericalCuttingPlaneGenerator:
    """Class that generates cutting planes with a spherical anchor point distribution.

    The normal can be either parallel to the anchor point or point in a random
    direction. The former results in cuts that mainly affect the surface whereas
    the latter results in completely random cuts.
    """

    def __init__(self, max_radius, min_radius=0.0, center=0.0, normal_dir='parallel'):
        self.center = center
        self.min_radius = min_radius
        self.max_radius = max_radius
        self.normal_dir = normal_dir

    def set_center(self, center):
        """Center around the anchor points will be distributed.

        Parameters:
            center : np.array of shape (3)
        """
        self.center = center

    def set_max_radius(self, max_radius):
        """Radius of the distribution where anchor point can be generated.

        Parameters:
            max_radius : float
        """
        self.max_radius = max_radius

    def generate_new_cutting_plane(self):
        """Generate a cutting plane by randomly choosing a normal and an anchor point."""
        normal = 1 - 2*np.random.random(3)
        normal /= np.linalg.norm(normal)

        if self.normal_dir == 'parallel':
            anchor_dir = normal
        else:
            anchor_dir = 1 - 2 * np.random.random(3)
            anchor_dir /= np.linalg.norm(anchor_dir)
        anchor = anchor_dir * (self.min_radius + np.random.random() * (self.max_radius - self.min_radius))
        anchor = anchor + self.center

        return CuttingPlane(anchor, normal)



import numpy as np


# TODO these features are experimental right now
class Population:
    def __init__(self):
        pass

    def add_offspring(self, particle):
        raise NotImplementedError

    def compute_fitness(self, particle):
        raise NotImplementedError

    def random_selection(self, n_individuals):
        raise NotImplementedError

    def tournament_selection(self, n_individuals, tournament_size):
        raise NotImplementedError

    def gaussian_tournament(self, n_individuals, tournament_size, mean=None):
        raise NotImplementedError


class NichedPopulation(Population):
    def __init__(self, niche_symbol, n_atoms):
        Population.__init__(self)

        self.niche_symbol = niche_symbol
        self.n_niches = n_atoms + 1
        self.population = dict()

    def __getitem__(self, item):
        return self.population[item]

    def compute_fitness(self, particle):
        niche = self.get_niche(particle)

        #return particle.get_energy('BRR')
        return 1

    def get_niche(self, particle):
        niche = particle.get_n_atoms_of_symbol(self.niche_symbol)

        return niche

    def add_offspring(self, particle):
        niche = self.get_niche(particle)

        self.population[niche] = particle

    def random_selection(self, n_individuals):
        return np.random.choice(list(self.population.values()), n_individuals, replace=False).tolist()

    def tournament_selection(self, n_individuals, tournament_size):
        winners = list()
        for tournament in range(n_individuals):
            candidates = np.random.choice(list(set(self.population.values()).difference(set(winners))), tournament_size, replace=False).tolist()
            candidates.sort(key=lambda x: self.compute_fitness(x))
            winners.append(candidates[0])

        return winners

    def gaussian_tournament(self, n_individuals, tournament_size, mean=None):
        winners = list()
        for tournament in range(n_individuals):
            if mean is None:
                mean = int(np.random.random() * self.n_niches)

            sigma = 5

            probabilities = np.array([1 / (sigma * 2 * np.pi) * np.exp(-0.5 * ((i - mean) / sigma) ** 2) for i in range(0, self.n_niches)])
            probabilities = probabilities / np.sum(probabilities)

            candidates = np.random.choice(list(self.population.values()), tournament_size, replace=False, p=probabilities).tolist()
            candidates.sort(key=lambda x: self.compute_fitness(x))

            winners.append(candidates[0])

        return winners

    def get_as_list(self):
        return list(self.population.values())

    def get_convex_hull(self):
        return [(niche, self.population[niche].get_energy('BRR')) for niche in range(self.n_niches)]





from ase import Atoms
from ase.io import read, write
import pickle
from npl.core.atom_wrapper import AtomWrapper
from npl.core.neighbor_list import NeighborList
from npl.core.adsorption import AdsorptionSiteList
from ase import Atoms
from ase.io import read, write


# TODO update local environment handling with keys
class BaseNanoparticle:
    """Represent a nanoparticle.

    Geometric information is stored in the form of an AtomWrapper object and
    a NeighborList object. Multiple energies, feature vectors and lcoal environments
    can be accesed by storing them with their respective key.

    BaseNanoparticle is a dataholder class in the sense that all common calculations
    such as the calculation of feature vectors, atomic descriptors or energies should
    be done by calculator classes that obtain the necessary data from BaseNanoparticle.
    This makes sure that BaseNanoparticle remains as general as possible.

    For project specific behavior the Nanoparticle class, which inherits from
    BaseNanoparticle should be used.
    """

    def __init__(self):
        self.atoms = AtomWrapper()
        self.neighbor_list = NeighborList()
        self.adsorption_site_list = AdsorptionSiteList()

        self.energies = dict()
        
        self.local_environments = dict()
        self.atom_features = dict()
        self.feature_vectors = dict()

    def get_geometrical_data(self):
        """Return the geometrical information of the nanoparticle as dictionary.

        This function returns a dict containing positions and the neighbor list, but not the
        symbols. In that way, nanoparticles with different orderings can share the same
        geometry.
        """
        data = dict()
        data['neighbor_list'] = self.neighbor_list.list
        data['positions'] = self.atoms.get_positions()

        return data

    def get_as_dictionary(self, fields=None):
        """Return the nanoparticle in the form of a JSON-compatible dictionary.

        Parameters:

            fields: list of str
                Specifies which piece of information the user wants to save. Accepted values are:
                'energies', 'symbols', 'positions', 'atom_features', 'local_environments',
                'neighbor_list', 'feature_vectors'.

        """

        full_particle_dict = {'energies': self.energies,
                              'symbols': list(self.atoms.get_symbols()),
                              'positions': self.atoms.get_positions(),
                              'atom_features': self.atom_features,
                              'local environments': self.local_environments,
                              'neighbor_list': dict(self.neighbor_list.list),
                              'feature_vectors': self.feature_vectors}

        if fields is None:
            return full_particle_dict
        else:
            data = {}
            for field in fields:
                data[field] = full_particle_dict[field]
            return data

    def save_npl_format(self, filename, fields, filename_geometry=None):
        """Save the nanoparticle by first exporting it to a dictionary and then using pickle.

        The nanoparticle can be stored into two separate files, one of which holds the geometry
        (i.e. positions and neighbor list, not the symbols!). This allows to reuse the same
        geometry for several particles with different ordering.

        Parameters:

            filename: str
            filename.

            fields: list of str
            Specify the piece of information to save. See get_as_dictionary() for explanation.

            filename_geometry: str
            Filename of file in which the geometrical information will be stored. This is independent
            of the fields variable.
        """
        data = self.get_as_dictionary(fields)
        pickle.dump(data, open(filename, 'wb'))

        if filename_geometry is not None:
            geometrical_data = self.get_geometrical_data()
            pickle.dump(geometrical_data, open(filename_geometry, 'wb'))

    def build_from_dictionary(self, particle_dict, geometrical_dict=None):
        """Construct a nanoparticle based on a representation as dictionary.

        Takes either one dictionary for the whole nanoparticle or two if one only specifies
        the geometry.

        Parameter:

            particle_dict: dict
            Dictionary containing all the necessary information to build a nanoparticle. See
            get_as_dictionary() for valid keys.

            geometrical_dict: dict
            Dictionary containing positions and neighbor list. If particle_dict also contains
            geometrical information, geometrical_dict will still be used.
        """
        if geometrical_dict is None:
            positions = particle_dict['positions']
            symbols = particle_dict['symbols']
        else:
            positions = geometrical_dict['positions']
            if particle_dict is None:
                symbols = ['X'] * len(positions)
            else:
                symbols = particle_dict['symbols']

        atoms = Atoms(symbols, positions)
        self.atoms.add_atoms(atoms)

        if particle_dict is None:
            self.neighbor_list.list = geometrical_dict['neighbor_list']
        else:
            if 'neighbor_list' in particle_dict:
                self.neighbor_list.list = particle_dict['neighbor_list']

        if particle_dict is not None:
            if 'neighbor_list' in particle_dict:
                self.neighbor_list.list = particle_dict['neighbor_list']
        if geometrical_dict is not None:
            if 'neighbor_list' in geometrical_dict:
                self.neighbor_list.list = geometrical_dict['neighbor_list']

        if particle_dict is not None:
            self.energies = particle_dict['energies']

            if 'feature_vectors' in particle_dict:
                self.feature_vectors = particle_dict['feature_vectors']
            if 'atom_features' in particle_dict:
                self.atom_features = particle_dict['atom_features']

            if 'local_environments' in particle_dict:
                self.local_environments = particle_dict['local_environments']

    def load_npl_format(self, filename, filename_geometry=None):
        """Load a nanoparticle file in the NPL format.

        Files are loaded using pickle.

        Parameter:

            filename: str
            Filename of the particle.

            filename_geometry: str
            Filename of geometrical data.
        """
        if filename is not None:
            dictionary = pickle.load(open(filename, 'rb'))
        else:
            dictionary = None

        if filename_geometry is not None:
            topological_data = pickle.load(open(filename_geometry, 'rb'))
            self.build_from_dictionary(dictionary, topological_data)
        else:
            self.build_from_dictionary(dictionary)

    def read(self, filename, construct_neighbor_list=True, energy_key=None):
        """Wrapper class around ase.io.read.

        By default a neighbor list will be constructed. Energies that are present in
        the atoms object, e.g. from after a DFT calculation can be given an individual
        energy key.

        Parameters:
            filename: str
            construct_neighbor_list: bool
            energy_key : str
        """
        atoms = read(filename)
        self.atoms.add_atoms(atoms)

        if construct_neighbor_list:
            self.construct_neighbor_list()

        if energy_key is not None:
            self.set_energy(energy_key, atoms.get_potential_energy())

    def write(self, filename):
        """Wrapper around the ase.io.write method.

        NPL specific information will NOT be stored, including energies

        Parameters:
            filename: str
        """
        atoms = self.atoms.get_ase_atoms()
        write(filename, atoms)

    def add_atoms(self, atoms, recompute_neighbor_list=True):
        """Add atoms to the nanoparticle.

        Neighbor list will be recomputed after the addition by default.

        Parameters:

            atoms: Atoms
            recompute_neighbor_list: bool
        """
        self.atoms.add_atoms(atoms)

        if recompute_neighbor_list:
            self.construct_neighbor_list()

    def remove_atoms(self, atom_indices, recompute_neighbor_list=True):
        """Remove atoms from the nanoparticle.

        Neighbor list will be recomputed after the removal by default.

        Parameters:

            atom_indices: list/array of int
            Indices of the atoms to be removed.

            recompute_neighbor_list: bool
        """
        self.atoms.remove_atoms(atom_indices)

        if recompute_neighbor_list:
            self.construct_neighbor_list()

    def swap_symbols(self, index_pairs):
        """Swap the elements of the specified pairs.

        Parameters:

            index_pairs: list of tuple
            List that contains tuples which contain two indices, respectively, of atoms where the symbols
            will be exchanged.
        """
        self.atoms.swap_symbols(index_pairs)

    def transform_atoms(self, atom_indices, new_symbols):
        """Change the symbol for the given atoms.

        The first atom will be changed to the first new symbol, the second atom to the second
        new symbol etc.

        Parameters:

            atom_indices: list/array of int

            new_symbols: list of str
        """
        self.atoms.transform_atoms(atom_indices, new_symbols)

    def random_ordering(self, stoichiometry):
        """Create a random chemical ordering for the particle with the given stoichiometry.

        Parameters:
            stoichiometry: dict
            Dictionary containing key-value pairs of the form symbol - number of atoms. Fractions
            are also supported, but care for round-off errors.

        Example:
              random_ordering({'Pt' : 79} \n
              random_ordering({'Pt' : 0.5, 'Au' : 0.5})
        """
        # account for stoichiometries given as proportions instead of absolute numbers
        if sum(stoichiometry.values()) == 1:
            n_atoms = self.atoms.get_n_atoms()
            transformed_stoichiometry = dict()
            for symbol in sorted(stoichiometry):
                transformed_stoichiometry[symbol] = int(n_atoms * stoichiometry[symbol])

            # adjust for round-off error
            if sum(transformed_stoichiometry.values()) != n_atoms:
                diff = n_atoms - sum(transformed_stoichiometry.values())
                transformed_stoichiometry[sorted(stoichiometry)[0]] += diff

            #print('Resulting stoichiometry: {}'.format(transformed_stoichiometry))
            self.atoms.random_ordering(transformed_stoichiometry)
        else:
            self.atoms.random_ordering(stoichiometry)

    def translate_atoms_positions(self, position):
        """ Shifts the origin of the coordinates towards a given position.
        
        Useful to center the origin on an atom, when the position passed
        are the coordinates of the atom.
        
        Parameters:
            position: array
            Array that contains x y z coordinates. """
        self.atoms.translate_atoms_positions(position)

    def get_indices(self):
        """Convenience function for range(n_atoms)."""
        return self.atoms.get_indices()

    def get_n_bonds(self):
        """Return the number of bonds, requires valid neighbor list."""
        return self.neighbor_list.get_n_bonds()

    def get_all_symbols(self):
        """Return list of symbols that occur at least once in the particle"""
        return self.atoms.get_all_symbols()

    def get_symbol(self, index):
        """Return symbol of the given index."""
        return self.atoms.get_symbol(index)

    def get_symbols(self, indices=None):
        """Return the elements in order of the passed indices.

        By default symbols for all atoms are returned.

        Parameters:
            indices : list/array of int
        """
        return self.atoms.get_symbols(indices)

    def get_indices_by_symbol(self, symbol):
        """Return indices of all atoms of the respective symbol.

        Parameters:
            symbol: str
        """
        return self.atoms.get_indices_by_symbol(symbol)

    def construct_neighbor_list(self, exclude_x=True, scale_factor=1.0):
        """Construct neighbor list.

        Vacancies denoted by symbol 'X' are excluded by the default.
        """
        stripped_atoms = self.get_ase_atoms(exclude_x=exclude_x)
        self.neighbor_list.construct(stripped_atoms, scale_factor=scale_factor)

    def get_atom_indices_from_coordination_number(self, coordination_numbers, symbol=None):
        """Return atom indices of atoms with certain coordination numbers.

        In addition, the search can be restricted to a spcific symbol.

        Parameters:
            coordination_numbers : list/array of int
            symbol : str
        """
        if symbol is None:
            return list(
                filter(lambda x: self.get_coordination_number(x) in coordination_numbers, self.atoms.get_indices()))
        else:
            return list(filter(lambda x: self.get_coordination_number(x) in coordination_numbers
                        and self.atoms.get_symbol(x) == symbol, self.atoms.get_indices()))

    def get_coordination_number(self, atom_idx):
        return self.neighbor_list.get_coordination_number(atom_idx)

    def get_coordination_atoms(self, atom_idx):
        return self.neighbor_list.get_coordination_atoms(atom_idx)

    def get_generalized_coordination_number(self, indices):
        return self.neighbor_list.get_generalized_coordination_number(indices)
    
    def get_generalized_coordination_numbers(self, sites):

        gcn_dict = dict()
        for site in sites:
            gen_cn = self.get_generalized_coordination_number(site)
            try:
                gcn_dict[gen_cn].append(site)
            except KeyError:
                gcn_dict[gen_cn] = [site]
        return gcn_dict


    def get_atoms_in_the_surface_plane(self, atom_idx, edges_corner=False):
        return self.neighbor_list.get_atoms_in_the_surface_plane(atom_idx, edges_corner=edges_corner)

    def get_n_atoms(self):
        """Return the number of atoms."""
        return self.atoms.get_n_atoms()

    def get_neighbor_list(self):
        return self.neighbor_list

    def get_ase_atoms(self, indices=None, exclude_x=True):
        """Return an ase Atoms object representing the nanoparticle.

        By default all atoms will be returned except vacancies.

        Parameters:
            indices: list/array of int
            Only the selected atoms will be returned. If None is given, all indices will be returned by default.

            exclude_x: bool
            Exclude vacancies denoted by symbol 'X' from the atoms object. Necessary
            e.g. for energy calculations
        """
        if exclude_x and 'X' in self.get_stoichiometry():
            vacancies = self.get_indices_by_symbol('X')
            if indices is None:
                valid_indices = set(self.get_indices()).difference(vacancies)
            else:
                valid_indices = set(indices).difference(vacancies)

            return self.atoms.get_ase_atoms(list(valid_indices))
        else:
            return self.atoms.get_ase_atoms(indices)

    def get_atoms(self, indices):
        return self.atoms.get_atoms(indices)

    def get_position(self, idx):
        return self.atoms.get_position(idx)

    def get_stoichiometry(self):
        return self.atoms.get_stoichiometry()

    def get_n_atoms_of_symbol(self, symbol):
        return self.atoms.get_n_atoms_of_symbol(symbol)

    def set_energy(self, energy_key, energy):
        self.energies[energy_key] = energy

    def get_energy(self, energy_key):
        return self.energies[energy_key]

    def has_energy(self, energy_key):
        if energy_key in self.energies:
            return True
        return False

    def set_feature_vector(self, feature_key, feature_vector):
        self.feature_vectors[feature_key] = feature_vector

    def get_feature_vector(self, feature_key):
        return self.feature_vectors[feature_key]

    def set_atom_features(self, atom_features, feature_key):
        self.atom_features[feature_key] = atom_features

    def set_atom_feature(self, feature_key, index, atom_feature):
        self.atom_features[feature_key][index] = atom_feature

    def get_atom_feature(self, feature_key, index):
        return self.atom_features[feature_key][index]

    def get_atom_features(self, feature_key):
        if feature_key not in self.atom_features:
            self.atom_features[feature_key] = dict()
        return self.atom_features[feature_key]

    def set_local_environment(self, atom_idx, local_environment):
        self.local_environments[atom_idx] = local_environment

    def get_local_environment(self, atom_idx):
        return self.local_environments[atom_idx]

    def set_local_environments(self, local_environments):
        self.local_environments = local_environments

    def get_local_environments(self):
        return self.local_environments

    def is_pure(self):
        if len(self.get_all_symbols()) == 1:
            return True
        return False

    ### ADSORPTION SECTION

    def construct_adsorption_list(self):
        self.adsorption_site_list.construct(self)

    def get_adsorption_list(self):
        return self.adsorption_site_list

    def get_adsorption_as_list(self):
        n_sites = self.get_total_number_of_sites()
        adsorption_site_list = [list(self.adsorption_site_list[x]) for x in range(n_sites)]
        return adsorption_site_list

    def get_total_number_of_sites(self):
        return self.adsorption_site_list.get_total_number_of_sites()

    def get_site_atom_indices(self, index):
        return self.adsorption_site_list.get_site_atom_indices(index)

    def random_occupation(self, number_of_adsorbates):
        self.adsorption_site_list.random_occupation(number_of_adsorbates)
        
    def get_occupation_vector(self):
        return self.adsorption_site_list.get_occupation_vector()

    def get_occupation_status_by_indices(self, status):
        return self.adsorption_site_list.get_occupation_status_by_indices(status)

    def get_number_of_adsorbates(self):
        return self.adsorption_site_list.get_number_of_adsorbates()

    def get_indices_of_adsorbates(self):
        return self.adsorption_site_list.get_occupation_status_by_indices(1)

    def swap_status(self, index_pairs):
        self.adsorption_site_list.swap_status(index_pairs)
        




import numpy as np
import copy


class ExchangeOperator:
    def __init__(self, p):
        self.p = p
        self.index = 0

    def random_exchange(self, particle, n_exchanges=None):
        new_particle = copy.deepcopy(particle)
        if new_particle.is_pure():
            print("Pure particle! No permutation possible")
            return new_particle

        symbols = new_particle.get_all_symbols()
        symbol1 = symbols[0]
        symbol2 = symbols[1]

        max_exchanges = min(new_particle.get_n_atoms_of_symbol(symbol1), new_particle.get_n_atoms_of_symbol(symbol2))
        if n_exchanges is None:
            n_exchanges = min(np.random.geometric(p=self.p, size=1)[0], max_exchanges)

        symbol1_indices = np.random.choice(new_particle.get_indices_by_symbol(symbol1), n_exchanges, replace=False)
        symbol2_indices = np.random.choice(new_particle.get_indices_by_symbol(symbol2), n_exchanges, replace=False)

        new_particle.swap_symbols(zip(symbol1_indices, symbol2_indices))

        return new_particle



import numpy as np
import copy

from npl.monte_carlo.random_exchange_operator_etop import RandomExchangeOperatorExtended

def setup_monte_carlo(start_particle, energy_calculator, feature_classifier):
    symbols = start_particle.get_all_symbols()
    energy_key = energy_calculator.get_energy_key()

    feature_calculator = feature_classifier

    feature_calculator.compute_feature_vector(start_particle)
    energy_calculator.compute_energy(start_particle)

    exchange_operator = RandomExchangeOperatorExtended(0.5)
    exchange_operator.bind_particle(start_particle)

    return energy_key, feature_calculator, exchange_operator

def features_to_update(start_particle, exchanges):

    neighborhood = set()
    for exchange in exchanges:
        index1, index2 = exchange
        neighborhood.add(index1)
        neighborhood.add(index2)

        neighborhood = neighborhood.union(start_particle.neighbor_list[index1])
        neighborhood = neighborhood.union(start_particle.neighbor_list[index2])

    return neighborhood

def run_monte_carlo(beta, max_steps, start_particle, energy_calculator, feature_classifier):
    energy_key, feature_calculator, exchange_operator = setup_monte_carlo(start_particle, energy_calculator, feature_classifier)

    start_energy = start_particle.get_energy(energy_key)
    lowest_energy = start_energy
    accepted_energies = [(lowest_energy, 0)]
    accepted_structures = []

    found_new_solution = False
    fields = ['energies', 'symbols', 'positions']
    best_particle = copy.deepcopy(start_particle.get_ase_atoms())

    total_steps = 0
    no_improvement = 0
    while no_improvement < max_steps:
        total_steps += 1
        if total_steps % 2000 == 0:
            print("Step: {}".format(total_steps))
            print("Lowest energy: {}".format(lowest_energy))

        exchanges = exchange_operator.random_exchange(start_particle)
        neighborhood = features_to_update(start_particle, exchanges)
        
        old_atom_features, change = feature_calculator.update_feature_vector(start_particle, neighborhood)

        energy_calculator.compute_energy(start_particle)
        new_energy = start_particle.get_energy(energy_key)

        delta_e = new_energy - start_energy

        acceptance_rate = min(1, np.exp(-beta * delta_e))
        if np.random.random() < acceptance_rate:
            if found_new_solution:
                if new_energy > start_energy:
                    start_particle.swap_symbols(exchanges)
                    best_particle = copy.deepcopy(start_particle.get_ase_atoms())
                    # accepted_structures.append(copy.deepcopy(start_particle.get_ase_atoms()))
                    # best_particle['energies'][energy_key] = start_energy
                    start_particle.swap_symbols(exchanges)

            start_energy = new_energy
            accepted_energies.append((new_energy, total_steps))
            
            if new_energy < lowest_energy:
                no_improvement = 0
                lowest_energy = new_energy
                found_new_solution = True
            else:
                no_improvement += 1
                found_new_solution = False

        else:
            no_improvement += 1

            # roll back exchanges and make sure features and environments are up-to-date
            start_particle.swap_symbols(exchanges)
            start_particle.set_energy(energy_key, start_energy)
            feature_calculator.downgrade_feature_vector(start_particle, neighborhood, old_atom_features,change)

            if found_new_solution:
                best_particle = copy.deepcopy(start_particle.get_ase_atoms())
                # accepted_structures.append(copy.deepcopy(start_particle.get_ase_atoms()))
                # best_particle['energies'][energy_key] = copy.deepcopy(start_energy)
                found_new_solution = False

    accepted_energies.append((accepted_energies[-1][0], total_steps))

    return [best_particle, accepted_energies]



from npl.monte_carlo.random_exchange_operator import RandomExchangeOperator

import numpy as np
import itertools

class RandomExchangeOperatorExtended(RandomExchangeOperator):
    def __init__(self, p_geometric):
        self.symbols = None
        self.n_symbols_atoms = 0
        
        self.swap_types = 0
        self.n_swap_types = None
        self.swap_types_probability = None
        
        self.max_exchanges = 0
        self.p_geometric = p_geometric
        
    def bind_particle(self, particle):
        self.symbols = sorted(particle.atoms.get_all_symbols())
        
        symbols_indices = dict()
        for symbol in self.symbols:
            symbols_indices[symbol] = particle.get_indices_by_symbol(symbol)
        
        
        self.max_exchanges = min([len(x) for x in symbols_indices.values()])
        
        self.swap_types = [swap_type for swap_type in itertools.combinations(self.symbols, 2)]
        self.get_swap_porbability(symbols_indices)
        self.n_symbols_atoms = [len(symbols_indices[symbol]) for symbol in self.symbols]
        self.n_swap_types = len(self.swap_types)

    def get_swap_porbability(self, symbols_indices):
        self.swap_types_probability = np.zeros(len(self.swap_types))
        for i, swap_type in enumerate(self.swap_types):
            self.swap_types_probability[i] += len(symbols_indices[swap_type[0]])
            self.swap_types_probability[i] += len(symbols_indices[swap_type[1]])

        self.swap_types_probability = self.swap_types_probability/sum(self.swap_types_probability)        
    
    def random_exchange(self, particle):
        swap_type_probability = np.random.choice(self.n_swap_types, 1, p=self.swap_types_probability)[0]
        symbol1, symbol2 = self.swap_types[swap_type_probability]
        max_exchanges = 1
        n_exchanges = min(np.random.geometric(p=self.p_geometric, size=1)[0], max_exchanges)
        symbol1_indices = np.random.choice(particle.get_indices_by_symbol(symbol1), n_exchanges, replace=False)
        symbol2_indices = np.random.choice(particle.get_indices_by_symbol(symbol2), n_exchanges, replace=False)

        particle.swap_symbols(list(zip(symbol1_indices, symbol2_indices)))
        return list(zip(symbol1_indices, symbol2_indices))



from npl.descriptors.local_environment_calculator import NeighborCountingEnvironmentCalculator
from npl.optimization.local_optimization.guided_exchange_operator import GuidedExchangeOperator
from npl.descriptors.local_environment_feature_classifier import TopologicalEnvironmentClassifier

import numpy as np

def setup_local_optimization(start_particle, energy_calculator, environment_energies, local_feature_classifier=None):
    symbols = start_particle.get_all_symbols()
    local_env_calculator = NeighborCountingEnvironmentCalculator(symbols)
    if local_feature_classifier is None:
        local_feature_classifier = TopologicalEnvironmentClassifier(local_env_calculator, symbols)

    local_env_calculator.compute_local_environments(start_particle)
    local_feature_classifier.compute_feature_vector(start_particle)

    feature_key = local_feature_classifier.get_feature_key()
    energy_calculator.compute_energy(start_particle)

    exchange_operator = GuidedExchangeOperator(environment_energies, feature_key)
    exchange_operator.bind_particle(start_particle)

    energy_key = energy_calculator.get_energy_key()

    return energy_key, local_env_calculator, local_feature_classifier, exchange_operator


def update_atomic_features(index1, index2, local_env_calculator, local_feature_classifier, particle):
    neighborhood = {index1, index2}
    neighborhood = neighborhood.union(particle.neighbor_list[index1])
    neighborhood = neighborhood.union(particle.neighbor_list[index2])
    
    for x in neighborhood:
        neighborhood = neighborhood.union(particle.neighbor_list[x])

    for index in neighborhood:
        local_env_calculator.compute_local_environment(particle, index)
        local_feature_classifier.compute_atom_feature(particle, index)

    local_feature_classifier.compute_feature_vector(particle, recompute_atom_features=False)

    return particle, neighborhood


def local_optimization(start_particle, energy_calculator, environment_energies, local_feature_classifier=None):
    energy_key, local_env_calculator, local_feature_classifier, exchange_operator = setup_local_optimization(
        start_particle, energy_calculator, environment_energies, local_feature_classifier)

    start_energy = start_particle.get_energy(energy_key)
    accepted_energies = [(start_energy, 0)]

    step = 0
    while True:
        step += 2
        index1, index2 = exchange_operator.guided_exchange(start_particle)
        exchanged_indices = [index1, index2]

        start_particle, neighborhood = update_atomic_features(index1, index2, local_env_calculator,
                                                              local_feature_classifier, start_particle)
        exchange_operator.update(start_particle, neighborhood, exchanged_indices)

        energy_calculator.compute_energy(start_particle)
        new_energy = start_particle.get_energy(energy_key)

        if new_energy < start_energy:
            start_energy = new_energy
            accepted_energies.append((new_energy, step))
        else:
            accepted_energies.append((start_energy, step))

            # roll back last exchange and make sure features and environments are up-to-date
            start_particle.swap_symbols([(index1, index2)])
            start_particle.set_energy(energy_key, start_energy)
            for index in neighborhood:
                local_env_calculator.compute_local_environment(start_particle, index)
                local_feature_classifier.compute_atom_feature(start_particle, index)

            break

    return [start_particle, accepted_energies]



from npl.optimization.local_optimization import setup_local_optimization
from npl.optimization.local_optimization import update_atomic_features
from npl.optimization.local_optimization.guided_exchange_operator import GuidedExchangeOperator

import copy
import time

def run_basin_hopping(start_particle, energy_calculator, environment_energies, n_hopping_attempts, n_hops,
                      local_feature_classifier=None):
    energy_key, local_env_calculator, local_feature_classifier, exchange_operator = setup_local_optimization(
        start_particle, energy_calculator, environment_energies, local_feature_classifier)

    start_energy = start_particle.get_energy(energy_key)
    lowest_energies = [(start_energy, 0)]
    best_particle = copy.deepcopy(start_particle)
    lowest_energy = start_energy
    times = []
    flip_energy_list = []

    step = 0
    for i in range(n_hopping_attempts):
        while True:
            step += 1
            index1, index2 = exchange_operator.guided_exchange(start_particle)

            flip = exchange_operator.symbol1_exchange_energies[index1] + exchange_operator.symbol2_exchange_energies[index2]

            exchanged_indices = [index1, index2]

            start_particle, neighborhood = update_atomic_features(index1, index2, local_env_calculator,
                                                                  local_feature_classifier, start_particle)

            exchange_operator.update(start_particle, neighborhood, exchanged_indices)


            energy_calculator.compute_energy(start_particle)
            new_energy = start_particle.get_energy(energy_key)

            flip_energy_list.append((flip, new_energy-start_energy, exchanged_indices))

            if new_energy < start_energy:
                start_energy = new_energy
                if new_energy < lowest_energy:
                    lowest_energy = new_energy
                    lowest_energies.append((lowest_energy, step))

            else:
                if i % 20 == 0:
                    print('Energy after local_opt: {:.3f}, lowest {:.3f}'.format(start_energy, lowest_energy))

                if lowest_energy == start_energy:
                    start_particle.swap_symbols([(index1, index2)])
                    best_particle = copy.deepcopy(start_particle)
                    best_particle.set_energy(energy_key, start_energy)

                    start_particle.swap_symbols([(index1, index2)])
                break

        for hop in range(n_hops):
            step += 1
            
            index1, index2 = exchange_operator.basin_hop_step(start_particle)

            exchanged_indices = [index1, index2]

            start_particle, neighborhood = update_atomic_features(index1, index2, local_env_calculator,
                                                                  local_feature_classifier, start_particle)
            exchange_operator.update(start_particle, neighborhood, exchanged_indices)
            
            energy_calculator.compute_energy(start_particle)
            new_energy = start_particle.get_energy(energy_key)

            start_energy = new_energy
    print('Lowest energy: {:.3f}'.format(lowest_energy))
    lowest_energies.append((lowest_energy, step))

    return [best_particle, lowest_energies, flip_energy_list]



from npl.core.nanoparticle import Nanoparticle
from npl.calculators import EMTCalculator
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import learning_curve
from sklearn.metrics import mean_absolute_error,mean_squared_error
from sklearn.linear_model import BayesianRidge,Ridge
from sklearn.model_selection import ShuffleSplit
import numpy as np

def plot_learning_curves(X, y, n_atoms, estimator, n_splits=10, train_sizes=range(1, 401, 10),y_lim=(0,)):
    """
    Plots learning curves for a given estimator using Mean Absolute Error (MAE) as the scoring metric.

    Parameters:
    X (array-like): Feature matrix.
    y (array-like): Target vector.
    n_atoms (int): Number of atoms, used for normalizing the error.
    estimator (object): The estimator object implementing 'fit' and 'predict' methods.
    n_splits (int, optional): Number of re-shuffling & splitting iterations for cross-validation. Default is 10.
    train_sizes (iterable, optional): Relative or absolute numbers of training examples that will be used to generate the learning curve. Default is range(1, 401, 10).

    The function performs cross-validation to compute training and test scores, calculates the quartiles for the scores,
    and plots the learning curves with shaded areas representing the interquartile ranges.
    """
    cv = ShuffleSplit(n_splits=n_splits, train_size=train_sizes[-1], test_size=len(X) - train_sizes[-1], random_state=0)
    train_sizes, train_scores, test_scores = learning_curve(estimator, X=X, y=y, cv=cv, n_jobs=-1, train_sizes=train_sizes, scoring='neg_mean_absolute_error')

    train_scores = [np.quantile(train_scores, quartile, axis=1) / n_atoms * 1000 for quartile in [0.25, 0.50, 0.75]]
    test_scores = [np.quantile(test_scores, quartile, axis=1) / n_atoms * 1000 for quartile in [0.25, 0.50, 0.75]]

    train_q25, train_q50, train_q75 = train_scores
    test_q25, test_q50, test_q75 = test_scores

    plt.fill_between(train_sizes, -train_q25, -train_q75, alpha=0.3, label='Train IQR')
    plt.fill_between(train_sizes, -test_q25, -test_q75, alpha=0.3, label='Test IQR')
    plt.plot(train_sizes, -train_q50, '--', label='Train Median')
    plt.plot(train_sizes, -test_q50, '-', label='Test Median')
    plt.ylim(y_lim)
    plt.ylabel('MAE [meV / atom]')
    plt.xlabel('Training Set Size')
    plt.title('Learning Curves')
    plt.legend()
    plt.grid(True)
    plt.show()


def get_reference_structure(particle: Nanoparticle, ase = None) -> Nanoparticle:
    if type(particle) is not Nanoparticle:
        p = Nanoparticle()
        p.add_atoms(particle)
        particle = p
    old_symbols = deepcopy(particle.atoms.atoms.symbols)
    new_symbols = ['Pt' for _ in particle.get_indices()]
    particle.transform_atoms(particle.get_indices(), new_symbols)
    EMTCalculator(relax_atoms=True).compute_energy(particle)
    particle.construct_neighbor_list()
    particle.transform_atoms(particle.get_indices(), old_symbols)
    if ase:
        return particle.get_ase_atoms()
    return particle

def plot_cummulative_success_rate(energies: list, steps: list, figname: str = None):
    """
    Plots the cumulative success rate based on given energies and steps, and saves the plot to a file.
    Parameters:
    energies (list): A list of energy values.
    steps (list): A list of step values corresponding to the energies.
    figname (str): The filename to save the plot.
    The function sorts the energies and steps, calculates the cumulative success rate, and plots it.
    The plot is saved as an image file with the specified filename.
    """
    energies, steps = zip(*sorted(zip(energies, steps)))
    min_energy = min(energies)
    max_steps = max(steps)
    success_rate = np.zeros(max_steps)
    
    percent = 0
    index = 0
    for step, energy in zip(steps, energies):
        if energy == min_energy:
            success_rate[step:] += 100 / len(energies)
    
   
    plt.plot(range(len(success_rate)), success_rate)
    
    plt.ylim(0,100)
    plt.xlabel('Steps')
    plt.ylabel('Success Rate (%)')
    plt.title('Cumulative Success Rate')
    plt.grid(True)
    plt.show()
    
    if figname:
        plt.savefig(figname, dpi=200)


from ase.optimize import BFGS
from ase.calculators.emt import EMT

import numpy as np
import copy
import pickle
import sklearn.gaussian_process as gp
from sklearn.linear_model import BayesianRidge
from sklearn.metrics import mean_absolute_error, root_mean_squared_error


class EnergyCalculator:
    """Base class for an energy calculator.

    Valid implementations have to implement the compute_energy(particle) function.
    Energies are saved in the particle object with the key of the respective calculator.
    """
    def __init__(self):
        self.energy_key = None
        pass

    def compute_energy(self, particle):
        raise NotImplementedError

    def get_energy_key(self):
        return copy.deepcopy(self.energy_key)

    def set_energy_key(self, energy_key):
        self.energy_key = energy_key

    def save(self, name_file : str):
        with open(name_file, 'wb') as out:
            pickle.dump(self, out)

    @staticmethod
    def load(name_file):
        with open(name_file, 'rb') as calc:
            return pickle.load(calc)

class EMTCalculator(EnergyCalculator):
    """Wrapper class around the asap3 EMT calculator."""

    def __init__(self, fmax=0.01, steps=50, relax_atoms=False):
        EnergyCalculator.__init__(self)
        self.fmax = fmax
        self.steps = steps
        self.energy_key = 'EMT'
        self.relax_atoms = relax_atoms

    def compute_energy(self, particle):
        """Compute the energy using EMT.

        BFGS is used for relaxation. By default, the atoms are NOT relaxed, i.e. the
        geometry remains unchanged unless this is explicitly stated.

        Parameters:
            particle : Nanoparticle
            relax_atoms : bool
        """
        cell_width = 1e3
        cell_height = 1e3
        cell_length = 1e3

        atoms = particle.get_ase_atoms(exclude_x=True)
        if not self.relax_atoms:
            atoms = atoms.copy()

        #atoms.set_cell(np.array([[cell_width, 0, 0], [0, cell_length, 0], [0, 0, cell_height]]))
        atoms.set_calculator(EMT())
        dyn = BFGS(atoms, logfile=None)
        dyn.run(fmax=self.fmax, steps=self.steps)

        energy = atoms.get_potential_energy()
        particle.set_energy(self.energy_key, energy)


class GPRCalculator(EnergyCalculator):
    """Energy calculator using global feature vectors and Gaussian Process Regression."""

    def __init__(self, feature_key, kernel=None, alpha=0.01, normalize_y=True):
        EnergyCalculator.__init__(self)
        if kernel is None:
            self.kernel = gp.kernels.ConstantKernel(1., (1e-1, 1e3)) * gp.kernels.RBF(1., (1e-3, 1e3))
        else:
            self.kernel = kernel

        self.alpha = alpha
        self.normalize_y = normalize_y
        self.GPR = None
        self.energy_key = 'GPR'
        self.feature_key = feature_key

    def fit(self, training_set, energy_key):
        """Fit the GPR model.

        The feature vectors with key = self.feature_key will be used for feature vectors. The
        energy with the specified energy_key will be the target function.

        Parameters:
            training_set : list of Nanoparticles
            energy_key : str
        """
        feature_vectors = [p.get_feature_vector(self.feature_key) for p in training_set]
        energies = [p.get_energy(energy_key) for p in training_set]

        self.GPR = gp.GaussianProcessRegressor(kernel=self.kernel, n_restarts_optimizer=20, alpha=self.alpha,
                                               normalize_y=self.normalize_y)
        self.GPR.fit(feature_vectors, energies)

    def compute_energy(self, particle):
        """Compute the energy using GPR.

        Assumes that a feature vector with key=self.feature_key is present in the particle.

        Parameters:
            particle : Nanoparticle
        """
        energy = self.GPR.predict([particle.get_feature_vector(self.feature_key)])[0]
        particle.set_energy(self.energy_key, energy)


class MixingEnergyCalculator(EnergyCalculator):
    """Compute the mixing energy using an arbitrary energy model.

    For the original energy model it is assumed that all previous steps in the energy pipeline, e.g. calculation
    of local environment, feature vectors etc. has been carried out.
    """
    def __init__(self, base_calculator=None, mixing_parameters=None, recompute_energies=False):
        EnergyCalculator.__init__(self)

        if mixing_parameters is None:
            self.mixing_parameters = dict()
        else:
            self.mixing_parameters = mixing_parameters

        self.base_calculator = base_calculator
        self.recompute_energies = recompute_energies
        self.energy_key = 'Mixing Energy'

    def compute_mixing_parameters(self, particle, symbols):
        """Compute the energies for the pure particles of the given symbols as reference points.

        Parameters:
            particle : Nanoparticle
            symbols : list of str
        """
        for symbol in symbols:
            particle.random_ordering({symbol: 1.0})
            self.base_calculator.compute_energy(particle)
            self.mixing_parameters[symbol] = particle.get_energy('EMT')

    def compute_energy(self, particle):
        """Compute the mixing energy of the particle using the base energy model.

        If energies have been computed using the same energy model as the base calculator they are reused if
        self.recompute_energies == False

        Parameters:
            particle : Nanoparticle
        """
        if self.recompute_energies:
            self.base_calculator.compute_energy(particle)

        energy_key = self.base_calculator.get_energy_key()
        mixing_energy = particle.get_energy(energy_key)

        n_atoms = particle.atoms.get_n_atoms()

        for symbol in particle.get_stoichiometry():
            mixing_energy -= self.mixing_parameters[symbol] * particle.get_stoichiometry()[symbol] / n_atoms

        particle.set_energy(self.energy_key, mixing_energy)


class BayesianRRCalculator(EnergyCalculator):
    """Energy calculator using global feature vectors and Bayesian Ridge Regression."""

    def __init__(self, feature_key):
        EnergyCalculator.__init__(self)

        self.ridge = BayesianRidge(fit_intercept=False)
        
        self.energy_key = 'BRR'
        self.feature_key = feature_key

    def fit(self, training_set, energy_key, validation_set=None):
        """Fit the Bayesian Ridge Regression (BRR) model.
        
        Parameters:
        ----------
        training_set : list of Nanoparticles
            The dataset used for training the model.
        energy_key : str
            The key used to extract energy values from the nanoparticles.
        validation_set : float or list of Nanoparticles, optional
            If a float is provided, it represents the fraction of the training set to be used as the validation set.
            If a list is provided, it is used as the validation set directly. Default is None.
        
        Returns:
        -------
        None
        """
        if isinstance(validation_set, float):
            split_index = int(len(training_set) * validation_set)
            validation_set = training_set[split_index:]
            training_set = training_set[:split_index]
        
        feature_vectors = [p.get_feature_vector(self.feature_key) for p in training_set]
        energies = [p.get_energy(energy_key) for p in training_set]

        self.ridge.fit(feature_vectors, energies)

        if validation_set:
            self.validate(validation_set, energy_key)

    def validate(self, validation_set, energy_key):
        """Validate the Bayesian Ridge Regression (BRR) model.
        
        Parameters:
        ----------
        validation_set : list of Nanoparticles
            The dataset used for validating the model.
        energy_key : str
            The key used to extract energy values from the nanoparticles.
        
        Returns:
        -------
        None
        """
        pred_validation = [self.compute_energy(p) for p in validation_set]
        true_validation = [p.get_energy(energy_key) for p in validation_set]
        print('Mean Absolute error {:.4f} eV'.format(mean_absolute_error(true_validation, pred_validation)))
        print('Root Mean Square error {:.4f} eV'.format(root_mean_squared_error(true_validation, pred_validation)))
            
    def get_coefficients(self):
        return self.ridge.coef_

    def set_coefficients(self, new_coefficients):
        self.ridge.coef_ = new_coefficients

    def set_feature_key(self, feature_key):
        self.feature_key = feature_key
    
    def compute_energy(self, particle):
        """Compute the energy using BRR.

        Assumes that a feature vector with key=self.feature_key is present in the particle.

        Parameters:
            particle : Nanoparticle
        """
        #brr_energy = self.ridge.predict([particle.get_feature_vector(self.feature_key)])
        #brr_energy = np.dot(np.transpose(self.ridge.coef_), particle.get_feature_vector(self.feature_key))
        brr_energy = np.dot(np.transpose(self.ridge.coef_), particle.get_feature_vector(self.feature_key))
        particle.set_energy(self.energy_key, brr_energy)
        return brr_energy


class DipoleMomentCalculator:
    
    def __init__(self):
        self.total_dipole_moment = None
        self.dipole_moments = None
        self.environments = None
        

    def compute_dipole_moment(self, particle, charges = [1, -1]):
    
        symbols = particle.get_all_symbols()
        fake_charges = {symbols[0] : charges[0], symbols[1] : charges[1]}
        partial_charges = [fake_charges[symbol] for symbol in particle.get_symbols()]

        dipole_moments = []
        environments = []
        for central_atom_idx in particle.get_atom_indices_from_coordination_number([12]):
            particle.translate_atoms_positions(particle.get_position(central_atom_idx))
            dipole_moment = 0
            for atom_idx in particle.get_coordination_atoms(central_atom_idx):
                dipole_moment += partial_charges[atom_idx] * particle.get_position(atom_idx)

            dipole_moments.append(np.linalg.norm(dipole_moment))
            environments.append(particle.get_coordination_atoms(central_atom_idx))

        self.total_dipole_moment = np.average(dipole_moments)/particle.get_n_atoms()
        self.dipole_moments = dipole_moments
        self.environments = environments

    def get_total_dipole_moment(self):
        return self.total_dipole_moment
    
    def get_dipole_moments(self):
        return self.dipole_moments

    def get_environments(self):
        return self.environments


class LateralInteractionCalculator:
    
    def __init__(self):
        EnergyCalculator.__init__(self)
        self.interaction_matrix = None
        self.energy_key = 'Lateral Interaction'
        self.a = 6

    def construct_interatomic_potential_matrix(self, particle):
        def construct_adsorbate_grid(particle):
            from npl.core.adsorption import PlaceAddAtoms
            particle.construct_adsorption_list()
            n_sites = particle.get_total_number_of_sites()
            ads_site_list = particle.get_adsorption_list()
            ads_placer = PlaceAddAtoms(particle.get_all_symbols())
            ads_placer.bind_particle(particle)
            adsorbate_positions = [list(ads_site_list[site]) for site in range(n_sites)] 
            particle = ads_placer.place_add_atom(particle, 'O', adsorbate_positions)
            return particle

        def get_adsorbate_distance_matrix(particle, n_atoms_np):
            ase_atoms = particle.get_ase_atoms()
            adsorbate_all_distances = ase_atoms.get_all_distances()[n_atoms_np:]
            distance_matrix = np.array([row[n_atoms_np:] for row in adsorbate_all_distances])
            return distance_matrix

        n_atoms_np = particle.get_n_atoms()
        particle = construct_adsorbate_grid(particle)
        distance_matrix = get_adsorbate_distance_matrix(particle, n_atoms_np)
        interaction_matrix = np.zeros(distance_matrix.shape)
        interaction_matrix = self.a / distance_matrix**2

        dimension = len(interaction_matrix)
        for i in np.arange(dimension):
            interaction_matrix[i][i] = 0

        self.interaction_matrix = interaction_matrix

    def bind_grid(self, particle):
        particle_for_grid = copy.deepcopy(particle)
        self.construct_interatomic_potential_matrix(particle_for_grid)

    def compute_energy(self, particle):
        lateral_interaction = 0
        occupied_sites_indices = particle.get_occupation_status_by_indices(1)
        for idx, site_index in enumerate(occupied_sites_indices):
            for pair_index in occupied_sites_indices[idx:]:
                lateral_interaction += self.interaction_matrix[site_index][pair_index]
                
        particle.set_energy(self.energy_key, lateral_interaction)        

# TODO move to relevant file -> Basin Hopping, Local optimization
# TODO remove scaling factors from topological descriptors
def compute_coefficients_for_linear_topological_model(global_topological_coefficients, symbols, n_atoms):
    coordination_numbers = list(range(13))
    symbols_copy = copy.deepcopy(symbols)
    symbols_copy.sort()
    symbol_a = symbols_copy[0]
    print("Coef symbol_a: {}".format(symbol_a))

    e_aa_bond = global_topological_coefficients[0]/n_atoms
    e_bb_bond = global_topological_coefficients[1]/n_atoms
    e_ab_bond = global_topological_coefficients[2]/n_atoms

    coefficients = []
    total_energies = []
    for symbol in symbols_copy:
        for cn_number in coordination_numbers:
            for n_symbol_a_atoms in range(cn_number + 1):
                energy = 0

                if symbol == symbol_a:
                    energy += (global_topological_coefficients[3]*0.1)  # careful...
                    energy += (n_symbol_a_atoms*e_aa_bond/2)
                    energy += ((cn_number - n_symbol_a_atoms)*e_ab_bond/2)
                    energy += (global_topological_coefficients[4 + cn_number])

                    total_energy = energy
                    total_energy += n_symbol_a_atoms*e_aa_bond/2
                    total_energy += (cn_number - n_symbol_a_atoms)*e_ab_bond/2
                else:
                    energy += (n_symbol_a_atoms*e_ab_bond/2)
                    energy += ((cn_number - n_symbol_a_atoms)*e_bb_bond/2)

                    total_energy = energy
                    total_energy += n_symbol_a_atoms*e_ab_bond/2
                    total_energy += (cn_number - n_symbol_a_atoms)*e_bb_bond/2

                coefficients.append(energy)
                total_energies.append(total_energy)

    coefficients = np.array(coefficients)

    return coefficients, total_energies


def compute_coefficients_for_shape_optimization(global_topological_coefficients, symbols):
    coordination_numbers = list(range(13))
    symbols_copy = copy.deepcopy(symbols)
    symbols_copy.sort()

    e_aa_bond = global_topological_coefficients[0]

    coordination_energies_a = dict()
    for index, cn in enumerate(coordination_numbers):
        coordination_energies_a[cn] = global_topological_coefficients[index + 1]

    coefficients = []
    total_energies = []
    for cn_number in coordination_numbers:
        for n_symbol_a_atoms in range(cn_number + 1):
            energy = 0

            energy += (n_symbol_a_atoms*e_aa_bond/2)
            energy += (coordination_energies_a[cn_number])

            total_energy = energy + n_symbol_a_atoms*e_aa_bond/2

            coefficients.append(energy)
            total_energies.append(total_energy)

    coefficients += [0]*len(coefficients)
    total_energies += [0]*len(total_energies)

    coefficients = np.array(coefficients)

    return coefficients, total_energies



import numpy as np
from npl.optimization.genetic_algoritm.CutAndSpliceOperator import CutAndSpliceOperator
from npl.optimization.genetic_algoritm.ExchangeOperator import ExchangeOperator
from npl.optimization.genetic_algoritm.MutationOperator import MutationOperator

import copy
import pickle


def locate_convex_hull(start_population, unsuccessful_gens_for_convergence,
                       energy_calculator, local_env_calculator, local_feature_classifier):
    gens_since_last_success = 0
    symbols = ['Pt', 'Au']

    mutation_operator = MutationOperator(0.5, symbols)
    exchange_operator = ExchangeOperator(0.5)
    cut_and_splice_operator = CutAndSpliceOperator(0, 10)

    population = copy.deepcopy(start_population)
    for p in population.population.values():
        local_env_calculator.compute_local_environments(p)
        local_feature_classifier.compute_feature_vector(p)
        energy_calculator.compute_energy(p)

    energy_log = [population.get_convex_hull()]

    cur_generation = 0
    while gens_since_last_success < unsuccessful_gens_for_convergence:
        cur_generation += 1

        if cur_generation % 1000 == 0:
            energy_log.append(population.get_convex_hull())
            pickle.dump(energy_log, open("energy_log.pkl", 'wb'))

        p = np.random.random()
        if p < 0.6:
            # choose two parents for cut and splice
            parent1, parent2 = population.tournament_selection(2, 5)
            new_particle = cut_and_splice_operator.cut_and_splice(parent1, parent2, False)

        elif p < 0.8:
            # random exchange
            parent = population.random_selection(1)[0]
            new_particle = exchange_operator.random_exchange(parent)

        else:
            # random mutation
            parent = population.npl.optimization.genetic_algoritmussian_tournament(1, 5)[0]
            new_particle = mutation_operator.random_mutation(parent)

        # check that it is not a pure particle
        if new_particle.is_pure():
            continue

        # TODO insert local optimization
        local_env_calculator.compute_local_environments(new_particle)
        local_feature_classifier.compute_feature_vector(new_particle)
        energy_calculator.compute_energy(new_particle)

        successful_offspring = False
        niche = population.get_niche(new_particle)

        if new_particle.get_energy('BRR') < population[niche].get_energy('BRR'):
            print("success in generation: {}".format(cur_generation))
            print("New Energy: {}".format(new_particle.get_energy('BRR')))
            successful_offspring = True
            population.add_offspring(new_particle)

        # reset counters and log energy
        if successful_offspring:
            gens_since_last_success = 0
        else:
            gens_since_last_success += 1

    return [population, energy_log, cur_generation]



# npl/optimization/__init__.py

from .go_search import GOSearch, MCSearch, GASearch, GuidedSearch
# from .monte_carlo import run_monte_carlos
from .basin_hopping import run_basin_hopping
#rom .genetic_algoritm import run_genetic_algorithm
from .local_optimization.local_optimization import local_optimization


__all__ = [
    "GOSearch",
    "MCSearch",
    "GASearch",
    "GuidedSearch",
    "local_optimization"
]


from npl.descriptors import LocalEnvironmentCalculator as LEC
from npl.descriptors import LocalEnvironmentFeatureClassifier as LFC
from npl.calculators import EnergyCalculator as EC
from npl.descriptors import GlobalFeatureClassifier as GFC

import time


class GOSearch:
    def __init__(self, find_global_minimum, create_start_configuration):
        self.find_global_minimum = find_global_minimum
        self.create_start_configuration = create_start_configuration

    def fit_energy_expression(self, training_set, symbols):
        raise NotImplementedError

    def build_args_list_for_gm_search(self, args_gm, args_start):
        raise NotImplementedError

    def build_kwargs_list_for_gm_search(self, additional_kwargs):
        raise NotImplementedError

    def run_multiple_simulations(self, n_runs, args_gm=None, kwargs_gm=None, args_start=None):
        run_times = []
        results = []
        for i in range(n_runs):
            print("Run: {}".format(i))

            args = self.build_args_list_for_gm_search(args_gm, args_start)
            kwargs = self.build_kwargs_list_for_gm_search(kwargs_gm)

            start_time = time.perf_counter()
            result = self.find_global_minimum(*args, **kwargs)
            end_time = time.perf_counter()

            run_time = end_time - start_time
            print("Runtime: {}".format(run_time))

            run_times.append(run_time)
            results.append(result)

        return results, run_times


class MCSearch(GOSearch):
    def __init__(self, find_global_minimum, create_start_configuration, beta, n_steps):
        GOSearch.__init__(self, find_global_minimum, create_start_configuration)
        self.beta = beta
        self.n_steps = n_steps
        self.energy_calculator = None
        self.local_feature_classifier = None

    def fit_energy_expression(self, training_set, symbols, energy_key = 'EMT'):
        local_env_calculator = LEC.NeighborCountingEnvironmentCalculator(symbols)
        global_feature_classifier = GFC.TopologicalFeatureClassifier(symbols)

        self.energy_calculator = EC.BayesianRRCalculator(global_feature_classifier.get_feature_key())
        self.local_feature_classifier = LFC.TopologicalEnvironmentClassifier(local_env_calculator, symbols)

        for p in training_set:
            global_feature_classifier.compute_feature_vector(p)

        self.energy_calculator.fit(training_set, energy_key)

        n_atoms = sum(list(training_set[0].get_stoichiometry().values()))
        lin_coef = self.energy_calculator.get_coefficients()
        topological_coefficients, _ = EC.compute_coefficients_for_linear_topological_model(lin_coef, symbols, n_atoms)
        print(topological_coefficients)

        self.energy_calculator.set_coefficients(topological_coefficients)
        self.energy_calculator.set_feature_key(self.local_feature_classifier.get_feature_key())

        return

    def build_args_list_for_gm_search(self, additional_args, args_start):
        if args_start is None:
            start_config = self.create_start_configuration()
        else:
            start_config = self.create_start_configuration(*args_start)

        args = [self.beta, self.n_steps, start_config, self.energy_calculator, self.local_feature_classifier]
        if additional_args is not None:
            return args + additional_args
        else:
            return args

    def build_kwargs_list_for_gm_search(self, additional_kwargs):
        if additional_kwargs is not None:
            return additional_kwargs
        else:
            return dict()

    def set_beta(self, beta):
        self.beta = beta

    def set_n_steps(self, n_steps):
        self.n_steps = n_steps


class GASearch(GOSearch):
    def __init__(self, genetic_algorithm, create_start_population, unsuccessful_gens_for_convergence):
        GOSearch.__init__(self, genetic_algorithm, create_start_population)
        self.unsuccessful_steps_for_convergence = unsuccessful_gens_for_convergence

        self.local_env_calculator = None
        self.energy_calculator = None
        self.local_feature_classifier = None
        self.total_energies = None

    def fit_energy_expression(self, training_set, symbols):
        local_env_calculator = LEC.NeighborCountingEnvironmentCalculator(symbols)
        global_feature_classifier = GFC.TopologicalFeatureClassifier(symbols)

        self.energy_calculator = EC.BayesianRRCalculator(global_feature_classifier.get_feature_key())
        self.local_feature_classifier = LFC.TopologicalEnvironmentClassifier(local_env_calculator, symbols)
        self.local_env_calculator = local_env_calculator

        for p in training_set:
            global_feature_classifier.compute_feature_vector(p)

        self.energy_calculator.fit(training_set, 'EMT')

        n_atoms = sum(list(training_set[0].get_stoichiometry().values()))
        lin_coef = self.energy_calculator.get_coefficients()
        print(lin_coef)
        topological_coefficients, self.total_energies = EC.compute_coefficients_for_linear_topological_model(
            lin_coef, symbols, n_atoms)

        self.energy_calculator.set_coefficients(topological_coefficients)
        self.energy_calculator.set_feature_key(self.local_feature_classifier.get_feature_key())

        return

    def build_args_list_for_gm_search(self, additional_args, args_start):
        if args_start is None:
            start_population = self.create_start_configuration()
        else:
            start_population = self.create_start_configuration(*args_start)
        args = [start_population, self.unsuccessful_steps_for_convergence, self.energy_calculator,
                self.local_env_calculator, self.local_feature_classifier, self.total_energies]

        if additional_args is not None:
            return args + additional_args
        else:
            return args

    def build_kwargs_list_for_gm_search(self, additional_kwargs):
        if additional_kwargs is not None:
            return additional_kwargs
        else:
            return dict()


class GuidedSearch(GOSearch):
    def __init__(self, find_global_minimum, create_start_configuration):
        GOSearch.__init__(self, find_global_minimum, create_start_configuration)
        self.total_energies = None
        self.energy_calculator = None

    def fit_energy_expression(self, training_set, symbols, energy_key='EMT'):
        global_feature_classifier = GFC.testTopologicalFeatureClassifier(symbols)

        self.energy_calculator = EC.BayesianRRCalculator(global_feature_classifier.get_feature_key())

        for p in training_set:
            global_feature_classifier.compute_feature_vector(p)

        self.energy_calculator.fit(training_set, energy_key)

        n_atoms = sum(list(training_set[0].get_stoichiometry().values()))
        lin_coef = self.energy_calculator.get_coefficients()
        print(lin_coef)
        topological_coefficients, self.total_energies = EC.compute_coefficients_for_linear_topological_model(
            lin_coef, symbols, n_atoms)

        self.energy_calculator.set_coefficients(topological_coefficients)
        self.energy_calculator.set_feature_key('TEC')
        return

    def build_args_list_for_gm_search(self, additional_args, args_start):
        if args_start is None:
            start_config = self.create_start_configuration()
        else:
            start_config = self.create_start_configuration(*args_start)

        args = [start_config, self.energy_calculator, self.total_energies]
        if additional_args is not None:
            return args + additional_args
        else:
            return args

    def build_kwargs_list_for_gm_search(self, additional_kwargs):
        if additional_kwargs is not None:
            return additional_kwargs
        else:
            return dict()


import numpy as np
import copy
from scipy.special import sph_harm


class LocalEnvironmentCalculator:
    """Base class for local environment calculators.

    Intended use: Implementations should implement predict_local_environment() which calculates and returns
    the local environment of a single atom.
    """
    def __init__(self):
        pass

    def compute_local_environments(self, particle):
        for index in particle.get_indices():
            self.compute_local_environment(particle, index)

    def compute_local_environment(self, particle, atom_index):
        local_env = self.predict_local_environment(particle, atom_index)
        particle.set_local_environment(atom_index, local_env)

    def predict_local_environment(self, particle, lattice_index):
        raise NotImplementedError


class SOAPCalculator(LocalEnvironmentCalculator):
    """ Experimental class. Compute the spherical harmonics power spectrum for each atom and element.

    Based on the code provided by Dr. Johannes Margraf.
    See https://arxiv.org/abs/1901.10971v1 Ceriotti et al. for details, used convention etc.
    """
    def __init__(self, l_max):
        LocalEnvironmentCalculator.__init__(self)
        self.l_max = l_max

    def predict_local_environment(self, particle, lattice_index):
        def map_onto_unit_sphere(cartesian_coordinates):
            # different use of the scipy.special.sph_harm notation for phi and theta (opposite of Wikipedia)
            def angular_from_cartesian_coordinates(cartesian_coordinates):
                x = cartesian_coordinates[0]
                y = cartesian_coordinates[1]
                z = cartesian_coordinates[2]

                hxy = np.hypot(x, y)
                r = np.hypot(hxy, z)
                el = np.arctan2(z, hxy)
                az = np.arctan2(y, x)
                return np.abs(az + np.pi), np.abs(el + np.pi / 2.0)

            return list(map(lambda x: angular_from_cartesian_coordinates(x), cartesian_coordinates))

        def spherical_harmonics_expansion():
            """
            This functions takes the environment atoms surrounding a reference atom
            in an fcc lattice and returns the spherical harmonic coefficients of the expansion
            """

            neighbors = list(particle.neighbor_list[lattice_index])
            symbols = np.array([particle.get_symbol(index) for index in neighbors])
            cartesian_coordinates = particle.atoms.get_positions(neighbors)

            center_atom_position = particle.atoms.get_positions([lattice_index])[0]
            cartesian_coordinates = list(map(lambda x: x - center_atom_position, cartesian_coordinates))

            angular_coordinates = map_onto_unit_sphere(cartesian_coordinates)

            # The density of each species is expanded separately
            expansion_coefficients = []
            n_neighbors = len(symbols)
            for symbol in sorted(particle.get_all_symbols()):
                symbol_density = np.zeros(n_neighbors)
                symbol_density[np.where(symbols == symbol)] += 1
                c_lms_symbol = []
                for l in range(self.l_max + 1):
                    for m in range(-l, l + 1):
                        c_lm = 0.0
                        for i, point in enumerate(angular_coordinates):
                            c_lm += symbol_density[i] * np.conj(sph_harm(m, l, point[0], point[1]))
                        c_lms_symbol.append(c_lm)
                expansion_coefficients.append(c_lms_symbol)
            return expansion_coefficients

        sh_expansion_coefficients = spherical_harmonics_expansion()
        bond_parameters = []
        for symbol_index_1, symbol_1 in enumerate(sorted(particle.get_all_symbols())):
            for symbol_index_2, symbol_2 in enumerate(sorted(particle.get_all_symbols())):
                q_ls_symbol = []
                i = 0
                for l in range(self.l_max + 1):
                    q_l = 0
                    for m in range(-l, l + 1):
                        q_l += np.conj(sh_expansion_coefficients[symbol_index_1][i]) * sh_expansion_coefficients[symbol_index_2][i]
                        i += 1
                    q_ls_symbol.append(np.pi*np.sqrt(8.0/(2*l + 1)*q_l))
                bond_parameters.append(q_ls_symbol)

        bond_parameters = np.array(bond_parameters)
        bond_parameters = bond_parameters.real
        # return the bond parameters as one big feature vector
        bond_parameters = np.reshape(bond_parameters, bond_parameters.size)

        return bond_parameters


class NeighborCountingEnvironmentCalculator(LocalEnvironmentCalculator):
    """Calculate a local environment of the form [n_a, n_b], where n_a denotes the number of surrounding a atoms.

    Currently restricted to two elements which are ordered alphabetically. Requires a valid neighbor list.
    """
    def __init__(self, symbols):
        LocalEnvironmentCalculator.__init__(self)
        symbols_copy = copy.deepcopy(symbols)
        symbols_copy.sort()
        self.symbol_a = symbols_copy[0]
        self.symbol_b = symbols_copy[1]

    def predict_local_environment(self, particle, lattice_index):
        n_a_atoms = 0
        n_b_atoms = 0

        neighbors = particle.neighbor_list[lattice_index]
        for neighbor in neighbors:
            if particle.get_symbol(neighbor) == self.symbol_a:
                n_a_atoms += 1
            else:
                n_b_atoms += 1

        return np.array([n_a_atoms, n_b_atoms])






from ase.units import kB as boltzmann_constant
from ase import Atoms
from ase.io.trajectory import Trajectory
import numpy as np
import random
from .base_ensemble import BaseEnsemble

class CanonicalEnsemble(BaseEnsemble):
    """
    Represents a canonical ensemble for Monte Carlo simulations.

    Args:
        atoms (Atoms): The initial atomic configuration.
        calculator (Calculator): The calculator used to compute energies and forces.
        random_seed (int, optional): The random seed for the PRNG. Defaults to None.
        optimizer (Optimizer, optional): The optimizer used for relaxation. Defaults to None.
        fmax (float, optional): The maximum force tolerance for relaxation. Defaults to 0.1.
        temperature (float, optional): The temperature of the ensemble in Kelvin. Defaults to 300.
        steps (int, optional): The number of Monte Carlo steps to perform. Defaults to 100.
        op_list (OperatorList, optional): The list of operators for mutations. Defaults to None.
        constraints (Constraints, optional): The constraints applied to the system. Defaults to None.
        traj_file (str, optional): The trajectory file name. Defaults to 'traj_test.traj'.
        outfile (str, optional): The output file name. Defaults to 'outfile.out'.
        outfile_write_interval (int, optional): The interval at which to write to the output file. Defaults to 10.

    Attributes:
        lowest_energy (float): The lowest potential energy found during the simulation.
        atoms (Atoms): The current atomic configuration.
        constraints (Constraints): The constraints applied to the system.
        _temperature (float): The temperature of the ensemble in Kelvin.
        _calculator (Calculator): The calculator used to compute energies and forces.
        _optimizer (Optimizer): The optimizer used for relaxation.
        _fmax (float): The maximum force tolerance for relaxation.
        _steps (int): The number of Monte Carlo steps to perform.
        _op_list (OperatorList): The list of operators for mutations.
        _step (int): The current step number.
        _accepted_trials (int): The number of accepted trials.

    """

    def __init__(self,
                 atoms,
                 calculator,
                 random_seed=None,
                 optimizer=None,
                 fmax=0.1,
                 temperature=300, 
                 steps=100, 
                 op_list=None, 
                 constraints=None,
                 traj_file: str = 'traj_test.traj', 
                 outfile: str = 'outfile.out',
                 outfile_write_interval: int = 10) -> None:
        """
        Initializes a CanonicalEnsemble object.

        Args:
            atoms (Atoms): The initial atomic configuration.
            calculator (Calculator): The calculator used to compute energies and forces.
            random_seed (int, optional): The random seed for the PRNG. Defaults to None.
            optimizer (Optimizer, optional): The optimizer used for relaxation. Defaults to None.
            fmax (float, optional): The maximum force tolerance for relaxation. Defaults to 0.1.
            temperature (float, optional): The temperature of the ensemble in Kelvin. Defaults to 300.
            steps (int, optional): The number of Monte Carlo steps to perform. Defaults to 100.
            op_list (OperatorList, optional): The list of operators for mutations. Defaults to None.
            constraints (Constraints, optional): The constraints applied to the system. Defaults to None.
            traj_file (str, optional): The trajectory file name. Defaults to 'traj_test.traj'.
            outfile (str, optional): The output file name. Defaults to 'outfile.out'.
            outfile_write_interval (int, optional): The interval at which to write to the output file. Defaults to 10.

        Returns:
            None
        """

        super().__init__(structure=atoms, calculator=calculator, random_seed=random_seed, traj_file=traj_file,outfile=outfile,outfile_write_interval=outfile_write_interval)

        if random_seed is not None:
            random.seed(random_seed)

        self.lowest_energy = float('inf')  # Initialize to positive infinity
        self.atoms = atoms
        self.constraints = constraints
        self._temperature = temperature
        self._calculator = calculator
        self._optimizer = optimizer
        self._fmax = fmax
        self._steps = steps
        self._op_list = op_list

        self._step = 0
        self._accepted_trials = 0

    def _acceptance_condition(self, potential_diff: float) -> bool:
        """
        Determines whether to accept a trial move based on the potential energy difference and temperature.

        Args:
            potential_diff (float): The potential energy difference between the current and new configurations.

        Returns:
            bool: True if the trial move is accepted, False otherwise.
        """

        if potential_diff <= 0:
            return True
        elif self._temperature <= 1e-16:
            return False
        else:
            p = np.exp(-potential_diff / (boltzmann_constant * self._temperature))
            return p > self._next_random_number()
    
    def _next_random_number(self) -> float:
        """
        Returns the next random number from the PRNG.

        Returns:
            float: The next random number.
        """

        return random.random()

    def relax(self, atoms) -> Atoms:
        """
        Relaxes the atomic configuration using the specified optimizer.

        Args:
            atoms (Atoms): The atomic configuration to relax.

        Returns:
            Atoms: The relaxed atomic configuration.
        """

        atoms.info['key_value_pairs'] = {}
        atoms.calc = self._calculator
        opt = self._optimizer(atoms, logfile=None)
        opt.run(fmax=self._fmax)
        
        Epot = atoms.get_potential_energy()
        atoms.info['key_value_pairs']['potential_energy'] = Epot
        return atoms

    def do_mutations(self):
        """
        Performs mutations on the current atomic configuration.

        Returns:
            Atoms: The mutated atomic configuration.
        """

        new_atoms = self.atoms.copy()  # Use copy to avoid modifying the original
        for op_i in range(np.random.geometric(p=0.35, size=1)[0]):
            new_atoms.info['data'] = {'tag': None}
            new_atoms.info['confid'] = 1
            operation = self._op_list.get_operator()
            new_atoms, _ = operation.get_new_individual([new_atoms])
        if self.constraints:
            new_atoms.set_constraint(self.constraints)    
        new_atoms = self.relax(new_atoms)
        return new_atoms

    def trial_step(self):
        """
        Performs a trial step in the Monte Carlo simulation.

        Returns:
            int: 1 if the trial move is accepted, 0 otherwise.
        """

        new_atoms = self.do_mutations()
        potential_diff = new_atoms.info['key_value_pairs']['potential_energy'] - self.atoms.info['key_value_pairs']['potential_energy']
        
        if self._acceptance_condition(potential_diff):
            if new_atoms.info['key_value_pairs']['potential_energy'] < self.lowest_energy:
                self.lowest_energy = new_atoms.info['key_value_pairs']['potential_energy']
                self.write_traj_file(self.atoms)
            self.atoms = new_atoms
            return 1
        return 0

    def run(self):
        """
        Runs the Monte Carlo simulation.

        Returns:
            None
        """

        self.relax(self.atoms)
        self.lowest_energy = self.atoms.get_potential_energy()
        for _ in range(self._steps):
            accepted = self.trial_step()
            self._step += 1
            self._accepted_trials += accepted

            if self._step % self._outfile_write_interval == 0:
                self.write_outfile(self._step, self.lowest_energy)

